# -*- coding: utf-8 -*-
r"""
obspy.clients.filesystem.tsindex - EarthScope TSIndex Client and Indexer
========================================================================

The obspy.clients.filesystem.tsindex module includes a timeseries extraction
:class:`Client` class for a database created by the EarthScope
`mseedindex <https://github.com/EarthScope/mseedindex/>`_ program, as well as,
a :class:`Indexer` class to create a SQLite3 database, following the
EarthScope `tsindex database schema
<https://github.com/EarthScope/mseedindex/wiki/Database-Schema/>`_\.

:copyright:
    Nick Falco, Chad Trabant, EarthScope (former IRIS) Data Services, 2024
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)


.. contents:: Contents
    :local:
    :depth: 2

Client Usage
------------

The first step is always to initialize a client object.

.. highlight:: python

>>> from obspy.clients.filesystem.tsindex import Client
>>> from obspy.clients.filesystem.tests.test_tsindex \
...     import get_test_data_filepath
>>> import os
>>> # for this example get the file path to test data
>>> filepath = get_test_data_filepath()
>>> db_path = os.path.join(filepath, 'timeseries.sqlite')
>>> # create a new Client instance
>>> client = Client(db_path, datapath_replace=("^", filepath))

The example below uses the test SQLite tsindex database included with ObsPy to
illustrate how to do the following:

* Determine what data is available in the tsindex database using
  :meth:`~Client.get_availability_extent()`
  and :meth:`~Client.get_availability()`, as
  well as, the percentage of data available using
  :meth:`~Client.get_availability_percentage()`.
* Request available timeseries data using
  :meth:`~Client.get_waveforms()` and
  :meth:`~Client.get_waveforms_bulk()`.

Determining Data Availability
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

* :meth:`~Client.get_availability_extent()`:
  Returns a list of (network, station, location, channel, earliest, latest)
  tuples that represent the full extent of available data. This example
  retrieves from the very small obspy test tsindex database a list of all
  available ("BHZ") channel extents from the Global Seismograph Network
  ("IU") for all times.

>>> extents = client.get_availability_extent(network="IU", channel="BHZ")
>>> for extent in extents:
...     print("{0:<3} {1:<6} {2:<3} {3:<4} {4} {5}".format(*extent))
IU  ANMO   10  BHZ  2018-01-01T00:00:00.019500Z 2018-01-01T00:00:59.994536Z
IU  COLA   10  BHZ  2018-01-01T00:00:00.019500Z 2018-01-01T00:00:59.994538Z

* :meth:`~Client.get_availability()`: Works in the same way as
  :meth:`~Client.get_availability_extent()` but returns a list of (network,
  station, location, channel, starttime, endtime) tuples representing
  contiguous time spans for selected channels and time ranges.

* :meth:`~Client.get_availability_percentage()`:
  Returns the tuple(float, int) of percentage of available data
  (``0.0`` to ``1.0``) and number of gaps/overlaps. Availability percentage is
  calculated relative to the provided ``starttime`` and ``endtime``.

>>> from obspy import UTCDateTime
>>> avail_percentage = client.get_availability_percentage(
...     "IU", "ANMO", "10", "BHZ",
...     UTCDateTime(2018, 1, 1, 0, 0, 0, 19500),
...     UTCDateTime(2018, 1, 1, 0, 1, 57, 994536))
>>> print(avail_percentage)
(0.5083705674817509, 1)

Requesting Timeseries Data
^^^^^^^^^^^^^^^^^^^^^^^^^^

* :meth:`~Client.get_waveforms()`:
  This example illustrates how to request 1 second of available ("IU")
  timeseries data in the test tsindex database. Results are returned as a
  :class:`~obspy.core.stream.Stream` object. See the
  :meth:`~Client.get_waveforms_bulk()`
  method for information on how to make multiple requests at once.

>>> t = UTCDateTime("2018-01-01T00:00:00.019500")
>>> st = client.get_waveforms("IU", "*", "*", "BHZ", t, t + 1)
>>> st.plot()  # doctest: +SKIP

.. plot::

    from obspy import UTCDateTime
    from obspy.clients.filesystem.tsindex import Client
    from obspy.clients.filesystem.tests.test_tsindex \
        import get_test_data_filepath
    import os
    # for this example get the file path to test data
    filepath = get_test_data_filepath()
    db_path = os.path.join(filepath, 'timeseries.sqlite')
    # create a new Client instance
    client = Client(db_path, datapath_replace=("^", filepath))
    t = UTCDateTime("2018-01-01T00:00:00.019500")
    st = client.get_waveforms("IU", "*", "*", "BHZ", t, t + 1)
    st.plot()

Indexer Usage
-------------

The :class:`~Indexer` provides a high level
API for indexing a directory tree of miniSEED files using the EarthScope
`mseedindex <https://github.com/EarthScope/mseedindex/>`_ software.

An important feature of this module is the ability to index data files
in parallel, making it convenient for indexing large data sets of many
files.

Initialize an indexer object by supplying the root path to data to be indexed.

>>> from obspy.clients.filesystem.tsindex import Indexer
>>> from obspy.clients.filesystem.tests.test_tsindex \
...     import get_test_data_filepath
>>> # for this example get the file path to test data
>>> filepath = get_test_data_filepath()
>>> # create a new Indexer instance
>>> indexer = Indexer(filepath, filename_pattern='*.mseed')

Index a directory tree of miniSEED files by calling
:meth:`~Indexer.run`. By default this will
create a database called ``timeseries.sqlite`` in the current working
directory. The name of the index database can be changed by supplying the
``database`` parameter when instantiating the
:class:`~Indexer` object.

.. code-block:: python

  indexer.run()

"""

import copyreg
import datetime
import logging
import os
import sqlalchemy as sa
import subprocess
import types

from collections import namedtuple
from glob import glob
from multiprocessing import Pool
from os.path import relpath
from sqlalchemy.exc import ResourceClosedError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import QueuePool

from obspy import UTCDateTime
from obspy.clients.filesystem.miniseed import _MiniseedDataExtractor, \
    NoDataError
from obspy.clients.filesystem.db import _get_tsindex_table, \
    _get_tsindex_summary_table
from obspy.core.stream import Stream
from obspy.core.util.decorator import deprecated_keywords

logger = logging.getLogger('obspy.clients.filesystem.tsindex')


def _pickle_method(m):
    """
    Allows serializing of class and instance methods.
    """
    if m.im_self is None:
        return getattr, (m.im_class, m.im_func.func_name)
    else:
        return getattr, (m.im_self, m.im_func.func_name)


copyreg.pickle(types.MethodType, _pickle_method)


class Client(object):
    """
    Time series extraction client for EarthScope tsindex database schema.
    """
    @deprecated_keywords({"loglevel": None})
    def __init__(self, database, datapath_replace=None, loglevel=None):
        """
        Initializes the client.

        :type database: str or
            :class:`~TSIndexDatabaseHandler`
        :param database: Path to SQLite tsindex database or a
            TSIndexDatabaseHandler object
        :type datapath_replace: tuple(str, str)
        :param datapath_replace: A ``tuple(str, str)``, where any
            occurrence of the first value will be replaced with the second
            value in filename paths from the index.
        :param loglevel: DEPRECATED and without effect
        """
        # setup handler for database
        if isinstance(database, str):
            self.request_handler = TSIndexDatabaseHandler(
                os.path.normpath(database))
        elif isinstance(database, TSIndexDatabaseHandler):
            self.request_handler = database
        else:
            raise ValueError("database must be a string or "
                             "TSIndexDatabaseHandler object.")

        # Create and configure the data extraction
        self.data_extractor = _MiniseedDataExtractor(
            dp_replace=datapath_replace)

    def get_waveforms(self, network, station, location,
                      channel, starttime, endtime, merge=-1):
        """
        Query tsindex database and read miniSEED data from local indexed
        directory tree.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window.
        :type merge: int or None
        :param merge: Specifies, which merge operation should be performed
            on the stream before returning the data. Default (``-1``) means
            only a conservative cleanup merge is performed to merge seamless
            traces (e.g. when reading across day boundaries). See
            :meth:`Stream.merge() <obspy.core.stream.Stream.merge>` for
            details. If set to ``None`` (or ``False``) no merge operation at
            all will be performed.
        :rtype: :class:`~obspy.core.stream.Stream`
        :returns: A ObsPy :class:`~obspy.core.stream.Stream` object containing
            requested timeseries data.
        """
        query_rows = [(network, station, location,
                       channel, starttime, endtime)]
        return self._get_waveforms(query_rows, merge)

    def get_waveforms_bulk(self, query_rows, merge=-1):
        """
        Query tsindex database and read miniSEED data from local indexed
        directory tree using a bulk request.

        :type query_rows: list(tuple(str, str, str, str,
            :class:`~obspy.core.utcdatetime.UTCDateTime`,
            :class:`~obspy.core.utcdatetime.UTCDateTime`)
        :param query_rows: A list of tuples [(net, sta, loc, cha, starttime,
            endtime),...] containing information on what timeseries should be
            returned from the indexed archive.
            Wildcards '*' and '?' are supported.
        :type merge: int or None
        :param merge: Specifies, which merge operation should be performed
            on the stream before returning the data. Default (``-1``) means
            only a conservative cleanup merge is performed to merge seamless
            traces (e.g. when reading across day boundaries). See
            :meth:`Stream.merge() <obspy.core.stream.Stream.merge>` for
            details. If set to ``None`` (or ``False``) no merge operation at
            all will be performed.
        :rtype: :class:`~obspy.core.stream.Stream`
        :returns: A ObsPy :class:`~obspy.core.stream.Stream` object containing
            requested timeseries data.
        """
        return self._get_waveforms(query_rows, merge)

    def get_nslc(self, network=None, station=None, location=None,
                 channel=None, starttime=None, endtime=None):
        """
        Get a list of tuples [(net, sta, loc, cha),...] containing information
        on what streams are included in the tsindex database.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window.
        :rtype: list(tuple(str, str, str, str))
        :returns: A list of tuples [(network, station, location, channel)...]
            containing information on what streams are included in the tsindex
            database.
        """
        summary_rows = self._get_summary_rows(network, station, location,
                                              channel, starttime, endtime)

        nslc_list = []
        for row in summary_rows:
            nslc = (row.network, row.station, row.location, row.channel)
            nslc_list.append(nslc)
        nslc_list.sort()
        return nslc_list

    def get_availability_extent(self, network=None, station=None,
                                location=None, channel=None, starttime=None,
                                endtime=None):
        """
        Get a list of tuples [(network, station, location, channel,
        earliest, latest)] containing data extent info for time series
        included in the tsindex database.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window.
        :rtype: list(tuple(str, str, str, str,
            :class:`~obspy.core.utcdatetime.UTCDateTime`,
            :class:`~obspy.core.utcdatetime.UTCDateTime`))
        :returns: A list of tuples [(network, station, location, channel,
            earliest, latest)...] containing data extent info for time series
            included in the tsindex database.
        """
        summary_rows = self._get_summary_rows(network, station,
                                              location, channel,
                                              starttime, endtime)

        availability_extents = []
        for row in summary_rows:
            extent = (row.network, row.station, row.location, row.channel,
                      UTCDateTime(row.earliest), UTCDateTime(row.latest))
            availability_extents.append(extent)
        availability_extents.sort()
        return availability_extents

    def get_availability(self, network=None, station=None, location=None,
                         channel=None, starttime=None, endtime=None,
                         include_sample_rate=False,
                         merge_overlap=False):
        """
        Get a list of tuples [(network, station, location, channel,
        starttime, endtime),...] containing data availability info for
        time series included in the tsindex database.

        If ``include_sample_rate=True``, then a tuple containing the sample
        rate [(net, sta, loc, cha, start, end, sample_rate),...] is returned.

        If ``merge_overlap=True``, then all time spans that overlap are merged.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window.
        :type include_sample_rate: bool
        :param include_sample_rate: If ``include_sample_rate=True``, then
            a tuple containing the sample rate [(net, sta, loc, cha,
            start, end, sample_rate),...] is returned.
        :type merge_overlap: bool
        :param merge_overlap: If ``merge_overlap=True``, then all time
            spans that overlap are merged.
        :rtype: list(tuple(str, str, str, str,
            :class:`~obspy.core.utcdatetime.UTCDateTime`,
            :class:`~obspy.core.utcdatetime.UTCDateTime`))
        :returns: A list of tuples [(network, station, location, channel,
            earliest, latest)...] representing contiguous time spans for
            selected channels and time ranges.
        """

        tsindex_rows = self._get_tsindex_rows(network, station,
                                              location, channel,
                                              starttime, endtime)

        grouped_channels = {}
        for row in tsindex_rows:
            if include_sample_rate is True:
                # split on different sample rates when merging
                hash = "{}_{}_{}_{}_{}".format(row.network,
                                               row.station,
                                               row.location,
                                               row.channel,
                                               row.samplerate)
            else:
                # ignore sample rate when merging
                hash = "{}_{}_{}_{}".format(row.network,
                                            row.station,
                                            row.location,
                                            row.channel)
            timespans = self._create_timespans_list(row.timespans)

            if grouped_channels.get(hash) is not None:
                if row.samplerate not in \
                        grouped_channels[hash]["samplerates"]:
                    grouped_channels[hash]["samplerates"].append(
                                                            row.samplerate)
                grouped_channels[hash]["timespans"].extend(timespans)
            else:
                grouped_channels[hash] = {}
                grouped_channels[hash]["samplerates"] = [row.samplerate]
                grouped_channels[hash]["timespans"] = timespans

        # sort timespans
        for _, channel_group in grouped_channels.items():
            channel_group["timespans"].sort()

        # join timespans
        joined_avail_tuples = []
        for sncl, channel_group in grouped_channels.items():
            net, sta, loc, cha = sncl.split("_")[:4]
            samplerates = channel_group["samplerates"]
            timespans = channel_group["timespans"]
            avail_data = self._get_availability_from_timespans(
                                                      net,
                                                      sta,
                                                      loc,
                                                      cha,
                                                      samplerates,
                                                      include_sample_rate,
                                                      merge_overlap,
                                                      timespans
                                                      )
            # extend complete list of available data
            joined_avail_tuples.extend(avail_data)
        joined_avail_tuples.sort()
        return joined_avail_tuples

    def get_availability_percentage(self, network, station,
                                    location, channel,
                                    starttime, endtime):
        """
        Get percentage of available data.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
        :type location: str
        :param location: Location code of requested data (e.g. "").
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window.
        :rtype: tuple(float, int)
        :returns: Tuple of percentage of available data (``0.0`` to ``1.0``)
            and number of gaps/overlaps.
        """
        avail = self.get_availability(network, station,
                                      location, channel,
                                      starttime, endtime,
                                      merge_overlap=True)

        if not avail:
            return (0, 1)

        total_duration = endtime - starttime
        # sum up gaps in the middle
        gap_sum = 0
        gap_count = 0
        for idx, cur_ts in enumerate(avail[1:]):
            prev_ts = avail[idx]
            gap_count = gap_count + 1
            gap_sum += cur_ts[4] - prev_ts[5]

        # check if we have a gap at start or end
        earliest = min([ts[4] for ts in avail])
        latest = max([ts[5] for ts in avail])
        if earliest > starttime:
            gap_sum += earliest - starttime
            gap_count += 1
        if latest < endtime:
            gap_sum += endtime - latest
            gap_count += 1
        return (1 - (gap_sum / total_duration), gap_count)

    def has_data(self, network=None, station=None, location=None,
                 channel=None, starttime=None, endtime=None):
        """
        Return whether there is data for a specified network, station,
        location, channel, starttime, and endtime combination.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window. Defaults to
            minimum possible start date.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window. Defaults to maximum
            possible end date.
        :rtype: bool
        :returns: Returns ``True`` if there is data in the index for a given
            network, station, location, channel, starttime, endtime.
        """
        if starttime is None:
            starttime = UTCDateTime(year=datetime.MINYEAR,
                                    month=1,
                                    day=1)
        if endtime is None:
            endtime = UTCDateTime(year=datetime.MAXYEAR,
                                  month=12,
                                  day=31)
        avail_percentage = self.get_availability_percentage(network,
                                                            station,
                                                            location,
                                                            channel,
                                                            starttime,
                                                            endtime)[0]
        if avail_percentage > 0:
            return True
        else:
            return False

    def _get_summary_rows(self, network, station, location, channel,
                          starttime, endtime):
        """
        Return a list of tuples [(net, sta, loc, cha, earliest, latest),...]
        containing information found in the tsindex_summary table.

        Information about the tsindex_summary schema may be found in the
        `mseedindex wiki <https://github.com/EarthScope/mseedindex/wiki/\
        Database-Schema#suggested-time-series-summary-table>`_.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        """
        query_rows = [(network, station, location,
                       channel, starttime, endtime)]
        return self.request_handler._fetch_summary_rows(query_rows)

    def _get_waveforms(self, query_rows, merge=-1):
        """
        Query tsindex database and read miniSEED data from local
        indexed directory tree using a bulk request and return a
        :class:`~obspy.core.stream.Stream` object containing the requested
        timeseries data.

        :type query_rows: list(tuple(str, str, str, str,
            :class:`~obspy.core.utcdatetime.UTCDateTime`,
            :class:`~obspy.core.utcdatetime.UTCDateTime`))
        :param query_rows: A list of tuples [(net, sta, loc, cha, starttime,
            endtime),...] containing information on what timeseries should be
            returned from the indexed archive.
            Wildcards '*' and '?' are supported.
        :type merge: int or None
        :param merge: Specifies, which merge operation should be performed
            on the stream before returning the data. Default (``-1``) means
            only a conservative cleanup merge is performed to merge seamless
            traces (e.g. when reading across day boundaries). See
            :meth:`Stream.merge() <obspy.core.stream.Stream.merge>` for
            details. If set to ``None`` (or ``False``) no merge operation at
            all will be performed.
        """
        # Get the corresponding index DB entries
        index_rows = self.request_handler._fetch_index_rows(query_rows)

        total_bytes = 0
        src_bytes = {}

        logger.debug("Starting data return")
        st = Stream(traces=[])
        try:
            # Extract the data, writing each returned segment to the response
            for data_segment in self.data_extractor.extract_data(index_rows):
                bytes_written = data_segment.get_num_bytes()
                src_name = data_segment.get_src_name()
                if bytes_written > 0:
                    st_segment = data_segment.read_stream()
                    st += st_segment
                    total_bytes += bytes_written
                    src_bytes.setdefault(src_name, 0)
                    src_bytes[src_name] += bytes_written
        except NoDataError:
            logger.debug("No data matched selection")

        logger.debug("Wrote {} bytes".format(total_bytes))

        if merge:
            st.merge(merge)
        return st

    def _get_tsindex_rows(self, network, station, location, channel, starttime,
                          endtime):
        """
        Return a list of tuples [(net, sta, loc, cha, quality... etc.),...]
        containing information found in the tsindex table.

        Information about the tsindex schema may be found in the
        `mseedindex wiki <https://github.com/EarthScope/mseedindex/wiki/\
        Database-Schema#sqlite-schema-version-11>`_.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window.
        """
        query_rows = [(network, station, location,
                       channel, starttime, endtime)]
        return self.request_handler._fetch_index_rows(query_rows)

    def _get_availability_from_timespans(self, network, station,
                                         location, channel,
                                         samplerates,
                                         include_sample_rate,
                                         merge_overlap,
                                         timespans,
                                         _sncl_joined_avail_tuples=None):
        """
        Recurse over a list of timespans, joining adjacent timespans,
        and merging if merge_overlaps is ``True``.

        Returns a list of tuples (network, station, location, channel,
        earliest, latest) representing available data.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window.
        :type timespans: list
        :param timespans: List of timespan tuples.
        """
        if _sncl_joined_avail_tuples is None:
            _sncl_joined_avail_tuples = []

        sr = min(samplerates)
        if len(timespans) > 1:
            prev_ts = timespans.pop(0)
            cur_ts = timespans.pop(0)
            if merge_overlap is True and \
                    self._do_timespans_overlap(prev_ts, cur_ts) is True:
                # merge if overlapping timespans and merge_overlap
                # option is set to true
                merged_ts = self._create_timespan(prev_ts.earliest,
                                                  cur_ts.latest)
                timespans.insert(0, merged_ts)
                return self._get_availability_from_timespans(
                                                network, station,
                                                location, channel,
                                                samplerates,
                                                include_sample_rate,
                                                merge_overlap,
                                                timespans,
                                                _sncl_joined_avail_tuples)
            elif self._are_timespans_adjacent(prev_ts, cur_ts, sr, 0.5):
                # merge if timespans are next to each other within
                # a 0.5 sample tolerance
                merged_ts = self._create_timespan(prev_ts.earliest,
                                                  cur_ts.latest)
                timespans.insert(0, merged_ts)
                return self._get_availability_from_timespans(
                                                network, station,
                                                location, channel,
                                                samplerates,
                                                include_sample_rate,
                                                merge_overlap,
                                                timespans,
                                                _sncl_joined_avail_tuples)
            else:
                # timespan shouldn't be merged so add to list
                avail_tuple = self._create_avail_tuple(
                                          network, station,
                                          location, channel,
                                          prev_ts.earliest,
                                          prev_ts.latest,
                                          sr=sr if include_sample_rate
                                          else None)
                _sncl_joined_avail_tuples.append(avail_tuple)
                timespans.insert(0, cur_ts)
                return self._get_availability_from_timespans(
                                                network, station,
                                                location, channel,
                                                samplerates,
                                                include_sample_rate,
                                                merge_overlap,
                                                timespans,
                                                _sncl_joined_avail_tuples)
        else:
            # no other timespans to merge with
            cur_ts = timespans.pop(0)
            avail_tuple = self._create_avail_tuple(
                                          network, station,
                                          location, channel,
                                          cur_ts.earliest,
                                          cur_ts.latest,
                                          sr=sr if include_sample_rate
                                          else None)
            _sncl_joined_avail_tuples.append(avail_tuple)
        return _sncl_joined_avail_tuples

    def _are_timespans_adjacent(self, ts1, ts2, sample_rate, tolerance=0.5):
        """
        Checks whether or not two time span named tuples
        (e.g. NameTuple(earliest, latest)) are adjacent within
        a given tolerance.

        :type ts1: NamedTuple
        :param ts1: Earliest timespan.
        :type ts2: NamedTuple
        :param ts2: Latest timespan.
        :type sample_rate: int
        :param sample_rate: Sensor sample rate.
        :type tolerance: float
        :param tolerance: Tolerance in seconds to determine whether a adjacent
            timespan should be merged.
        """
        # @40Hz sample period = 0.025
        sample_period = 1. / float(sample_rate)
        expected_next = ts1.latest + sample_period
        # @40Hz tolerance = 0.0125
        tolerance_amount = (tolerance*sample_period)
        actual_next = ts2.earliest
        if expected_next + tolerance_amount > actual_next and \
           expected_next - tolerance_amount < actual_next:
            return True
        else:
            return False

    def _do_timespans_overlap(self, ts1, ts2):
        """
        Checks whether or not two time span named tuples
        (e.g. NameTuple(earliest, latest)) intersect with
        one another.

        :type ts1: :class:`collections.namedtuple`
        :param ts1: Earliest timespan.
        :type ts2: :class:`collections.namedtuple`
        :param ts2: Latest timespan.
        """
        if ts1.earliest <= ts2.latest and \
           ts1.latest >= ts2.earliest:
            return True
        else:
            return False

    def _create_avail_tuple(self, network, station, location, channel,
                            earliest, latest, sr=None):
        """
        Returns a tuple representing available waveform data.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type earliest: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param earliest: Earliest date of timespan.
        :type latest: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param latest: Latest date of timespan.
        :type sr: float
        :param sr: Sensor sample rate (optional).
        """
        if sr is not None:
            avail_record = (network, station, location, channel,
                            UTCDateTime(float(earliest)),
                            UTCDateTime(float(latest)), sr)
        else:
            avail_record = (network, station, location, channel,
                            UTCDateTime(float(earliest)),
                            UTCDateTime(float(latest)))
        return avail_record

    def _create_timespan(self, earliest, latest):
        """
        Create a ``TimeSpan`` named tuple object given a earliest and latest
        date.

        :param earliest: Earliest date of timespan.
        :param latest: Latest date of timespan.
        """
        TimeSpan = namedtuple('TimeSpan',
                              ['earliest', 'latest'])
        return TimeSpan(earliest, latest)

    def _create_timespans_list(self, raw_timespans):
        """
        Given a timespans string from the database, return
        a list of named tuples.

        :type raw_timespans: str
        :param raw_timespans: timespans field from tsindex database table.
        """
        timespans = []
        unparsed_timespans = \
            raw_timespans.replace("[", "").replace("]", "").split(",")
        for t in unparsed_timespans:
            earliest, latest = t.split(":")
            ts = self._create_timespan(float(earliest), float(latest))
            timespans.append(ts)
        return timespans


class Indexer(object):
    """
    Build an index for miniSEED data using EarthScope's mseedindex program.
    Recursively search for files matching ``filename_pattern`` starting
    from ``root_path`` and run ``index_cmd`` for each target file found that
    is not already in the index. After all new files are indexed a summary
    table is generated with the extents of each timeseries.
    """
    @deprecated_keywords({"leap_seconds_file": None})
    @deprecated_keywords({"loglevel": None})
    def __init__(self, root_path, database="timeseries.sqlite",
                 index_cmd='mseedindex',
                 bulk_params=None, filename_pattern='*', parallel=5,
                 leap_seconds_file=None, loglevel=None):
        """
        Initializes the Indexer.

        :type root_path: str
        :param root_path: Root path to the directory structure to index.
        :type database: str or
            :class:`~TSIndexDatabaseHandler`
        :param database: Path to SQLite tsindex database or a
            TSIndexDatabaseHandler object. A database will be created
            if one does not already exists at the specified path.
        :type index_cmd: str
        :param index_cmd: Command to be run for each target file found that
            is not already in the index
        :type bulk_params: dict
        :param bulk_params: Dictionary of options to pass to ``index_cmd``.
        :type filename_pattern: str
        :param filename_pattern: Glob pattern to determine what files to index.
        :type parallel: int
        :param parallel: Max number of ``index_cmd`` instances to run in
            parallel. By default a max of 5 parallel process are run.
        :param loglevel: DEPRECATED and without effect
        :param leap_seconds_file: DEPRECATED and without effect
        """
        self.index_cmd = index_cmd
        if bulk_params is None:
            bulk_params = {}
        self.bulk_params = bulk_params
        self.filename_pattern = filename_pattern
        self.parallel = parallel

        # setup handler for database
        if isinstance(database, str):
            self.request_handler = TSIndexDatabaseHandler(database)
        elif isinstance(database, TSIndexDatabaseHandler):
            self.request_handler = database
        else:
            raise ValueError("Database must be a string or "
                             "TSIndexDatabaseHandler object.")

        self.root_path = os.path.abspath(root_path)
        if not os.path.isdir(self.root_path):
            raise OSError("Root path `{}` does not exists."
                          .format(self.root_path))

    def run(self, build_summary=True, relative_paths=False, reindex=False):
        """
        Execute the file discovery and indexing.

        :type build_summary: bool
        :param build_summary: By default, a summary table is (re)generated
            containing the extents for each time series in the index. This can
            be turned off by setting `build_summary` to False.
        :type relative_paths: bool
        :param relative_paths: By default, the absolute path to each file is
            stored in the index. If ``relative_paths`` is True, the file paths
            will be relative to the ``root_path``.
        :type reindex: bool
        :param reindex: By default, files are not indexed that are already in
            the index and have not been modified.  The ``reindex`` option can
            be set to ``True`` to force a re-indexing of all files regardless.
        """
        if self._is_index_cmd_installed() is False:
            raise OSError(
                    "Required program '{}' is not installed. Hint: Install "
                    "mseedindex at https://github.com/EarthScope/mseedindex/."
                    .format(self.index_cmd))
        if self.request_handler.sqlite:
            self.request_handler._set_sqlite_pragma()

        try:
            file_paths = self.build_file_list(relative_paths, reindex)
        except OSError as error:
            print(error)
            return

        # always keep the original file paths as specified. absolute and
        # relative paths are determined in the build_file_list method
        self.bulk_params["-kp"] = None

        if self.bulk_params.get("-table") is None:
            self.bulk_params["-table"] = self.request_handler.tsindex_table

        if (self.bulk_params.get("-sqlite") is None and
                self.request_handler.sqlite and
                self.request_handler.database is not None):
            self.bulk_params['-sqlite'] = self.request_handler.database

        pool = Pool(processes=self.parallel)
        # run mseedindex on each file in parallel
        try:
            proccesses = []
            for file_name in file_paths:
                logger.debug("Indexing file '{}'.".format(file_name))
                proc = pool.apply_async(Indexer._run_index_command,
                                        args=(self.index_cmd,
                                              self.root_path,
                                              file_name,
                                              self.bulk_params))
                proccesses.append(proc)
            pool.close()
            # Without timeout, cannot respond to KeyboardInterrupt.
            # Also need get to raise the exceptions workers may throw.
            for proc in proccesses:
                cmd, rc, out, err = proc.get(timeout=999999)
                if rc:
                    logger.warning("FAIL [{0}] '{1}' out: '{2}' err: '{3}'"
                                   .format(rc, cmd, out, err))
            pool.join()
        except KeyboardInterrupt:
            logger.warning('Parent received keyboard interrupt.')
            if build_summary is True:
                logger.warning("Skipped building timeseries summary "
                               "table since indexing was ended "
                               "prematurely.")
            pool.terminate()
        else:
            if build_summary is True:
                self.request_handler.build_tsindex_summary()

    def build_file_list(self, relative_paths=False, reindex=False):
        """
        Create a list of absolute paths to all files under ``root_path`` that
        match the ``filename_pattern``.

        :type relative_paths: bool
        :param relative_paths: By default, the absolute path to each file is
            stored in the index. If ``relative_paths`` is ``True``, then the
            file paths will be relative to the ``root_path``.
        :type reindex: bool
        :param reindex: If ``reindex`` is ``True``, then already indexed
            files will be reindexed.
        :rtype: list(str)
        :returns: A list of files under the ``root_path`` matching
            ``filename_pattern``.
        """
        logger.debug("Building a list of files to index.")
        file_list = self._get_rootpath_files(relative_paths=False)
        # find relative file paths in case they are stored in the database as
        # relative paths.
        file_list_relative = self._get_rootpath_files(relative_paths=True)
        result = []
        if reindex is False and self.request_handler.has_tsindex():
            # remove any files already in the tsindex table
            unindexed_abs = []
            unindexed_rel = []
            tsindex = self.request_handler._fetch_index_rows()
            tsindex_filenames = [os.path.normpath(row.filename)
                                 for row in tsindex]
            for abs_fn, rel_fn in zip(file_list, file_list_relative):
                if abs_fn not in tsindex_filenames and \
                   rel_fn not in tsindex_filenames:
                    unindexed_abs.append(abs_fn)
                    unindexed_rel.append(rel_fn)
            if relative_paths is True:
                result = unindexed_rel
            else:
                result = unindexed_abs
        elif relative_paths is True:
            result = file_list_relative
        else:
            result = file_list
        if not result:
            raise OSError("No {}files matching filename pattern '{}' "
                          "were found under root path '{}'."
                          .format("unindexed " if reindex is False else "",
                                  self.filename_pattern,
                                  self.root_path))
        return result

    def _get_rootpath_files(self, relative_paths=False):
        """
        Return a list of absolute paths to files under the rootpath that
        match the Indexers filename pattern
        """
        file_list = [os.path.normpath(y) for x in os.walk(self.root_path)
                     for y in glob(os.path.join(x[0], self.filename_pattern))
                     if os.path.isfile(y)]
        if relative_paths:
            file_list_relative = []
            for abs_path in file_list:
                file_list_relative.append(os.path.normpath(relpath(
                    abs_path, self.root_path)))
            return file_list_relative
        else:
            return file_list

    def _is_index_cmd_installed(self):
        """
        Checks if the index command (e.g. mseedindex) is installed.

        :rtype: bool
        :returns: Returns ``True`` if the ``index_cmd`` is installed.
        """
        try:
            subprocess.call([self.index_cmd, "-V"])
        except OSError:
            return False
        else:
            return True

    @classmethod
    def _run_index_command(cls, index_cmd, root_path, file_name, bulk_params):
        """
        Execute a command to perform indexing.

        :type index_cmd: str
        :param index_cmd: Name of indexing command to execute. Defaults to
            ``mseedindex``.
        :type file_name: str
        :param file_name: Name of file to index.
        :type bulk_params: dict
        :param bulk_params: Dictionary of options to pass to ``index_cmd``.
        """
        try:
            cmd = [index_cmd]
            for option, value in bulk_params.items():
                params = [option, value]
                cmd.extend(params)
            cmd.append(file_name)
            # boolean options have a value of None
            cmd = [c for c in cmd if c is not None]
            proc = subprocess.Popen(cmd,
                                    cwd=root_path,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            out, err = proc.communicate()
            return (index_cmd, proc.returncode,
                    out.strip(), err.strip())
        except Exception as err:
            msg = ("Error running command `{}` - {}"
                   .format(index_cmd, err))
            raise OSError(msg)


class TSIndexDatabaseHandler(object):
    """Supports direct tsindex database data access and manipulation.

    .. warning:: Direct use of this class is experimental.  It cannot yet
    be used to support TSIndex operation on other databases.

    """

    @deprecated_keywords({"loglevel": None})
    def __init__(self, database=None, tsindex_table="tsindex",
                 tsindex_summary_table="tsindex_summary",
                 session=None, loglevel=None):
        """
        Main query interface to timeseries index database.

        :type database: str
        :param database: Path to sqlite tsindex database
        :type tsindex_table: str
        :param tsindex_table: Name of timeseries index table
        :type tsindex_summary_table: str
        :param tsindex_summary_table: Name of timeseries index summary table
        :type session: :class:`sqlalchemy.orm.session.Session`
        :param session: An existing database session object.
        """
        self.database = None
        self.tsindex_table = tsindex_table
        self.tsindex_summary_table = tsindex_summary_table
        self.TSIndexTable = _get_tsindex_table(self.tsindex_table)
        self.TSIndexSummaryTable = \
            _get_tsindex_summary_table(self.tsindex_summary_table)

        if database and session:
            raise ValueError("Both a database path and an existing database "
                             "session object were supplied. Supply one or "
                             "the other but not both.")

        if session:
            self.session = session
            self.engine = session().get_bind()
        elif database:
            if isinstance(database, str):
                self.database = os.path.abspath(database)
                db_dirpath = os.path.dirname(self.database)
                if not os.path.exists(db_dirpath):
                    raise OSError("Database path '{}' does not exist."
                                  .format(db_dirpath))
                elif not os.path.isfile(self.database):
                    logger.info("No SQLite database file exists at `{}`."
                                .format(self.database))
            else:
                raise ValueError("database must be a string.")
            db_path = "sqlite:///{}".format(self.database)
            self.engine = sa.create_engine(db_path, poolclass=QueuePool)
            self.session = sessionmaker(bind=self.engine)
        else:
            raise ValueError("Either a database path or an existing "
                             "database session object must be supplied.")

        self.sqlite = True if 'sqlite' in self.engine.dialect.name.lower() \
            else False

    def get_tsindex_summary_cte(self):
        """
        :rtype: `sqlalchemy.sql.expression.CTE`
        :returns: Returns a common table expression (CTE) containing the
            tsindex summary information. If a tsindex summary table has been
            created in the database it will be used as the source for the CTE,
            otherwise it will be created by querying the tsindex table directly
            as a, potentially slow, fallback method.
        """
        session = self.session()
        tsindex_summary_cte_name = "tsindex_summary_cte"
        if self.has_tsindex_summary():
            # get tsindex summary cte by querying tsindex_summary table
            tsindex_summary_cte = \
                (session
                 .query(self.TSIndexSummaryTable)
                 .group_by(self.TSIndexSummaryTable.network,
                           self.TSIndexSummaryTable.station,
                           self.TSIndexSummaryTable.location,
                           self.TSIndexSummaryTable.channel)
                 .cte(name=tsindex_summary_cte_name)
                 )
        else:
            logger.warning("No {0} table found! A {0} "
                           "CTE will be created by querying the {1} "
                           "table, which could be slow!"
                           .format(self.tsindex_summary_table,
                                   self.tsindex_table))
            logger.info("For improved performance create a permanent "
                        "{0} table by running the "
                        "TSIndexDatabaseHandler.build_tsindex_summary() "
                        "instance method."
                        .format(self.tsindex_summary_table))
            # create the tsindex summary cte by querying the tsindex table.
            tsindex_summary_cte = \
                (session
                 .query(self.TSIndexTable.network,
                        self.TSIndexTable.station,
                        self.TSIndexTable.location,
                        self.TSIndexTable.channel,
                        sa.func.min(self.TSIndexTable.starttime)
                        .label("earliest"),
                        sa.func.max(self.TSIndexTable.endtime)
                        .label("latest"),
                        sa.literal(
                            UTCDateTime.now().isoformat()).label("updt")
                        )
                 .group_by(self.TSIndexTable.network,
                           self.TSIndexTable.station,
                           self.TSIndexTable.location,
                           self.TSIndexTable.channel)
                 .cte(name=tsindex_summary_cte_name)
                 )
        return tsindex_summary_cte

    def build_tsindex_summary(self):
        """
        Builds a tsindex_summary table using the table name supplied to the
        Indexer instance (defaults to 'tsindex_summary').
        """
        # test if tsindex table exists
        if not self.has_tsindex():
            raise ValueError("No tsindex table '{}' exists in database '{}'."
                             .format(self.tsindex_table, self.database))
        if self.has_tsindex_summary():
            self.TSIndexSummaryTable.__table__.drop(self.engine)

        session = self.session()
        self.TSIndexSummaryTable.__table__.create(self.engine)
        rows = (session
                .query(self.TSIndexTable.network,
                       self.TSIndexTable.station,
                       self.TSIndexTable.location,
                       self.TSIndexTable.channel,
                       sa.func.min(self.TSIndexTable.starttime)
                       .label("earliest"),
                       sa.func.max(self.TSIndexTable.endtime)
                       .label("latest"),
                       sa.literal(UTCDateTime().now().isoformat()))
                .group_by(self.TSIndexTable.network,
                          self.TSIndexTable.station,
                          self.TSIndexTable.location,
                          self.TSIndexTable.channel))
        session.execute(self.TSIndexSummaryTable.__table__.insert(),
                        [{'network': r[0],
                          'station': r[1],
                          'location': r[2],
                          'channel': r[3],
                          'earliest': r[4],
                          'latest': r[5],
                          'updt': r[6]
                          }
                        for r in rows])
        session.commit()

    def has_tsindex_summary(self):
        """
        Returns ``True`` if there is a tsindex_summary table in the database.

        :rtype: bool
        :returns: Returns ``True`` if there a tsindex_summary table is present
            in the database.
        """
        table_names = sa.inspect(self.engine).get_table_names()
        if self.tsindex_summary_table in table_names:
            return True
        else:
            return False

    def has_tsindex(self):
        """
        Returns ``True`` if there is a tsindex table in the database.

        :rtype: bool
        :returns: Returns ``True`` if there a tsindex table is present
            in the database.
        """
        table_names = sa.inspect(self.engine).get_table_names()
        if self.tsindex_table in table_names:
            return True
        else:
            return False

    def _fetch_index_rows(self, query_rows=None, bulk_params=None):
        '''
        Fetch index rows matching specified request. This method is marked as
        private because the index schema is subject to change.

        :type query_rows: list(tuple(str, str, str, str,
            :class:`~obspy.core.utcdatetime.UTCDateTime`,
            :class:`~obspy.core.utcdatetime.UTCDateTime`))
        :param query_rows: List of tuples containing (net,sta,loc,chan,start,
            end). By default everything is selected.
        :type bulk_params: dict
        :param bulk_params: Dict of bulk parameters (e.g. quality)
            Request elements may contain '?' and '*' wildcards.  The start and
            end elements can be a single '*' if not a date-time string.
        :rtype: list(tuple)
        :returns: Return rows as list of named tuples containing:
            (network, station, location, channel, quality, starttime, endtime,
            samplerate, filename, byteoffset, bytes, hash, timeindex,
            timespans, timerates, format, filemodtime, updated, scanned,
            requeststart, requestend).
        '''

        session = self.session()

        if query_rows is None:
            query_rows = []
        if bulk_params is None:
            bulk_params = {}

        query_rows = self._clean_query_rows(query_rows)
        request_cte_name = "raw_request_cte"

        result = []
        # Create a CTE that contains the request
        try:
            stmts = [
                sa.select(
                    sa.literal(a).label("network"),
                    sa.literal(b).label("station"),
                    sa.literal(c).label("location"),
                    sa.literal(d).label("channel"),
                    sa.case((sa.literal(e) == '*',
                             sa.literal('0000-00-00T00:00:00')),
                            else_=sa.literal(e)
                            ).label("starttime"),
                    sa.case((sa.literal(f) == '*',
                             sa.literal('5000-00-00T00:00:00')),
                            else_=sa.literal(f)
                            ).label("endtime")
                )
                for idx, (a, b, c, d, e, f) in enumerate(query_rows)
            ]
            requests = sa.union_all(*stmts)
            requests_cte = requests.cte(name=request_cte_name)
            wildcards = False
            for req in query_rows:
                for field in req:
                    if '*' in str(field) or '?' in str(field):
                        wildcards = True
                        break
            summary_present = self.has_tsindex_summary()
            if wildcards and summary_present:
                # Resolve wildcards using summary if present to:
                # a) resolve wildcards, allows use of '=' operator
                #    and table index
                # b) reduce index table search to channels that are
                #    known included
                flattened_request_cte_name = 'flattened_request_cte'
                # expand
                flattened_request_cte = (
                    session
                    .query(self.TSIndexSummaryTable.network,
                           self.TSIndexSummaryTable.station,
                           self.TSIndexSummaryTable.location,
                           self.TSIndexSummaryTable.channel,
                           sa.case((requests_cte.c.starttime == '*',
                                    self.TSIndexSummaryTable.earliest),
                                   else_=requests_cte.c.starttime
                                   ).label('starttime'),
                           sa.case((requests_cte.c.endtime == '*',
                                    self.TSIndexSummaryTable.latest),
                                   else_=requests_cte.c.endtime
                                   ).label('endtime'))
                    .filter(self.TSIndexSummaryTable.network.op('GLOB')
                            (requests_cte.c.network))
                    .filter(self.TSIndexSummaryTable.station.op('GLOB')
                            (requests_cte.c.station))
                    .filter(self.TSIndexSummaryTable.location.op('GLOB')
                            (requests_cte.c.location))
                    .filter(self.TSIndexSummaryTable.channel.op('GLOB')
                            (requests_cte.c.channel))
                    .filter(self.TSIndexSummaryTable.earliest <=
                            requests_cte.c.endtime)
                    .filter(self.TSIndexSummaryTable.latest >=
                            requests_cte.c.starttime)
                    .order_by(self.TSIndexSummaryTable.network,
                              self.TSIndexSummaryTable.station,
                              self.TSIndexSummaryTable.location,
                              self.TSIndexSummaryTable.channel,
                              self.TSIndexSummaryTable.earliest,
                              self.TSIndexSummaryTable.latest)
                    .cte(name=flattened_request_cte_name))
                result = (
                    session
                    .query(self.TSIndexTable,
                           requests_cte.c.starttime,
                           requests_cte.c.endtime)
                    .filter(self.TSIndexTable.network ==
                            flattened_request_cte.c.network)
                    .filter(self.TSIndexTable.station ==
                            flattened_request_cte.c.station)
                    .filter(self.TSIndexTable.location ==
                            flattened_request_cte.c.location)
                    .filter(self.TSIndexTable.channel ==
                            flattened_request_cte.c.channel)
                    .filter(self.TSIndexTable.starttime <=
                            flattened_request_cte.c.endtime)
                    .filter(self.TSIndexTable.endtime >=
                            flattened_request_cte.c.starttime)
                    .order_by(self.TSIndexTable.network,
                              self.TSIndexTable.station,
                              self.TSIndexTable.location,
                              self.TSIndexTable.channel,
                              self.TSIndexTable.starttime,
                              self.TSIndexTable.endtime))
                wildcards = False
            else:
                result = (
                    session
                    .query(self.TSIndexTable,
                           requests_cte.c.starttime,
                           requests_cte.c.endtime
                           )
                    .filter(self.TSIndexTable.network.op('GLOB')
                            (requests_cte.c.network))
                    .filter(self.TSIndexTable.station.op('GLOB')
                            (requests_cte.c.station))
                    .filter(self.TSIndexTable.location.op('GLOB')
                            (requests_cte.c.location))
                    .filter(self.TSIndexTable.channel.op('GLOB')
                            (requests_cte.c.channel))
                    .filter(self.TSIndexTable.starttime <=
                            requests_cte.c.endtime)
                    .filter(self.TSIndexTable.endtime >=
                            requests_cte.c.starttime)
                    .order_by(self.TSIndexTable.network,
                              self.TSIndexTable.station,
                              self.TSIndexTable.location,
                              self.TSIndexTable.channel,
                              self.TSIndexTable.starttime,
                              self.TSIndexTable.endtime))
        except Exception as err:
            raise ValueError(str(err))

        index_rows = []
        try:
            for rt in result:
                # convert to a named tuple
                NamedRow = namedtuple('NamedRow',
                                      ['network', 'station', 'location',
                                       'channel', 'quality', 'version',
                                       'starttime', 'endtime', 'samplerate',
                                       'filename', 'byteoffset', 'bytes',
                                       'hash', 'timeindex', 'timespans',
                                       'timerates', 'format', 'filemodtime',
                                       'updated', 'scanned', 'requeststart',
                                       'requestend'])
                row, requeststart, requestend = rt
                nrow = NamedRow(
                        row.network, row.station, row.location,
                        row.channel, row.quality, row.version,
                        row.starttime, row.endtime, row.samplerate,
                        row.filename, row.byteoffset, row.bytes,
                        row.hash, row.timeindex, row.timespans,
                        row.timerates, row.format, row.filemodtime,
                        row.updated, row.scanned,
                        requeststart, requestend
                       )
                index_rows.append(nrow)
        except ResourceClosedError:
            pass  # query returned no results
        logger.debug("Fetched %d index rows" % len(index_rows))
        return index_rows

    def _fetch_summary_rows(self, query_rows):
        '''
        Fetch summary rows matching specified request. A temporary tsindex
        summary is created if one does not exists. This method is marked
        as private because the index schema is subject to change.

        Returns rows as list of named tuples containing:
        (network,station,location,channel,earliest,latest,updated)

        :type query_rows: list(tuple(str, str, str, str,
            :class:`~obspy.core.utcdatetime.UTCDateTime`,
            :class:`~obspy.core.utcdatetime.UTCDateTime`))
        :param query_rows: List of tuples containing
            (net,sta,loc,chan,start,end) Request elements may contain '?'
            and '*' wildcards. The start and end elements can be a single
            '*' if not a date-time string.
        :rtype: list(tuple)
        :returns: Return rows as list of named tuples containing:
            (network, station, location, channel, earliest, latest, updated).
        '''
        session = self.session()
        query_rows = self._clean_query_rows(query_rows)
        tsindex_summary_cte = self.get_tsindex_summary_cte()
        # Create a CTE that contains the request
        try:
            request_cte_name = "request_cte"
            stmts = [
                sa.select(
                    sa.literal(a).label("network"),
                    sa.literal(b).label("station"),
                    sa.literal(c).label("location"),
                    sa.literal(d).label("channel"),
                    sa.case((sa.literal(e) == '*',
                             sa.literal('0000-00-00T00:00:00')),
                            else_=sa.literal(e)
                            ).label("starttime"),
                    sa.case((sa.literal(f) == '*',
                             sa.literal('5000-00-00T00:00:00')),
                            else_=sa.literal(f)
                            ).label("endtime")
                    )
                for idx, (a, b, c, d, e, f) in enumerate(query_rows)
            ]
            requests = sa.union_all(*stmts)
            requests_cte = requests.cte(name=request_cte_name)
        except Exception as err:
            raise ValueError(str(err))

        # Select summary rows by joining with summary table
        try:
            # expand
            result = (
                session
                .query(tsindex_summary_cte.c.network,
                       tsindex_summary_cte.c.station,
                       tsindex_summary_cte.c.location,
                       tsindex_summary_cte.c.channel,
                       tsindex_summary_cte.c.earliest,
                       tsindex_summary_cte.c.latest,
                       tsindex_summary_cte.c.updt)
                .filter(tsindex_summary_cte.c.network.op('GLOB')
                        (requests_cte.c.network))
                .filter(tsindex_summary_cte.c.station.op('GLOB')
                        (requests_cte.c.station))
                .filter(tsindex_summary_cte.c.location.op('GLOB')
                        (requests_cte.c.location))
                .filter(tsindex_summary_cte.c.channel.op('GLOB')
                        (requests_cte.c.channel))
                .filter(tsindex_summary_cte.c.earliest <=
                        requests_cte.c.endtime)
                .filter(tsindex_summary_cte.c.latest >=
                        requests_cte.c.starttime)
                .order_by(tsindex_summary_cte.c.network,
                          tsindex_summary_cte.c.station,
                          tsindex_summary_cte.c.location,
                          tsindex_summary_cte.c.channel,
                          tsindex_summary_cte.c.earliest,
                          tsindex_summary_cte.c.latest))
        except Exception as err:
            raise ValueError(str(err))

        # Map raw tuples to named tuples for clear referencing
        NamedRow = namedtuple('NamedRow',
                              ['network', 'station', 'location', 'channel',
                               'earliest', 'latest', 'updated'])
        summary_rows = []
        try:
            for row in result:
                summary_rows.append(NamedRow(*row))
        except ResourceClosedError:
            pass  # query returned no results
        logger.debug("Fetched %d summary rows" % len(summary_rows))
        return summary_rows

    def _create_query_row(self, network, station, location,
                          channel, starttime, endtime):
        """
        Returns a tuple (network, station, location, channel, starttime,
        endtime) with elements that have been formatted to match database
        entries. This allows for accurate comparisons when querying the
        database.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        """
        # Replace "--" location ID request alias with true empty value
        if location == "--":
            location = ""
        if isinstance(starttime, UTCDateTime):
            starttime = starttime.isoformat()
        if isinstance(endtime, UTCDateTime):
            endtime = endtime.isoformat()
        return (network, station, location, channel, starttime, endtime)

    def _format_date(self, dt):
        try:
            if dt is None:
                dt = "*"
            else:
                dt = UTCDateTime(dt)
        except TypeError:
            raise TypeError("'{}' could not be converted to "
                            "type 'UTCDateTime'.".format(dt))
        return dt

    def _clean_query_rows(self, query_rows):
        """
        Reformats query rows to match what is stored in the database.

        :type query_rows: list(tuple(str, str, str, str,
            :class:`~obspy.core.utcdatetime.UTCDateTime`,
            :class:`~obspy.core.utcdatetime.UTCDateTime`))
        :param query_rows: List of tuples containing (network, station,
            location, channel, starttime, endtime).
        """
        flat_query_rows = []
        if query_rows == []:
            # if an empty list is supplied then select everything
            select_all_query = self._create_query_row('*', '*', '*',
                                                      '*', '*', '*')
            flat_query_rows = [select_all_query]
        else:
            # perform some formatting on the query rows to ensure that they
            # query the database properly.
            for i, qr in enumerate(query_rows):
                query_rows[i] = self._create_query_row(*qr)

            # flatten query rows
            for req in query_rows:
                networks = (req[0].replace(" ", "").split(",")
                            if req[0] else "*")
                stations = (req[1].replace(" ", "").split(",")
                            if req[1] else "*")
                locations = (req[2].replace(" ", "").split(",")
                             if req[2] else "*")
                channels = (req[3].replace(" ", "").split(",")
                            if req[3] else "*")
                starttime = self._format_date(req[4])
                endtime = self._format_date(req[5])

                for net in networks:
                    for sta in stations:
                        for loc in locations:
                            for cha in channels:
                                qr = self._create_query_row(net, sta, loc, cha,
                                                            starttime, endtime)
                                flat_query_rows.append(qr)
        return flat_query_rows

    def _set_sqlite_pragma(self):
        """
        Setup a SQLite database for indexing.
        """
        try:
            logger.debug('Setting up SQLite database {}'.
                         format(self.database if self.database else ""))
            # setup the sqlite database
            session = self.session()
            # https://www.sqlite.org/foreignkeys.html
            session.execute(sa.text('PRAGMA foreign_keys = ON'))
            # as used by mseedindex
            session.execute(sa.text('PRAGMA case_sensitive_like = ON'))
            # enable Write-Ahead Log for better concurrency support
            session.execute(sa.text('PRAGMA journal_mode=WAL'))
        except Exception:
            raise OSError("Failed to setup SQLite database for indexing.")


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
Header files for the FDSN webservice.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import platform
import sys

from obspy import UTCDateTime, __version__


class FDSNException(Exception):
    status_code = None

    def __init__(self, value, server_info=None):
        if server_info is not None:
            if self.status_code is None:
                value = "\n".join([value, "Detailed response of server:", "",
                                   server_info])
            else:
                value = "\n".join([value,
                                   "HTTP Status code: {}"
                                   .format(self.status_code),
                                   "Detailed response of server:",
                                   "",
                                   server_info])
        super(FDSNException, self).__init__(value)


class FDSNNoDataException(FDSNException):
    status_code = 204


class FDSNBadRequestException(FDSNException):
    status_code = 400


class FDSNUnauthorizedException(FDSNException):
    status_code = 401


class FDSNForbiddenException(FDSNException):
    status_code = 403


class FDSNRequestTooLargeException(FDSNException):
    status_code = 413


class FDSNTooManyRequestsException(FDSNException):
    status_code = 429


class FDSNInternalServerException(FDSNException):
    status_code = 500


class FDSNNotImplementedException(FDSNException):
    status_code = 501


class FDSNBadGatewayException(FDSNException):
    status_code = 502


class FDSNServiceUnavailableException(FDSNException):
    status_code = 503


class FDSNTimeoutException(FDSNException):
    pass


class FDSNRedirectException(FDSNException):
    pass


class FDSNNoAuthenticationServiceException(FDSNException):
    pass


class FDSNDoubleAuthenticationException(FDSNException):
    pass


class FDSNInvalidRequestException(FDSNException):
    pass


class FDSNNoServiceException(FDSNException):
    pass


# A curated list collecting FDSNWS implementations
# primarily but not exclusively drawn from:
#    https://www.fdsn.org/ws/datacenters/1/query
# last updated November 5 2025
URL_MAPPINGS = {
    "AUSPASS": "https://auspass.edu.au",
    "BGR": "https://eida.bgr.de",
    "BGS": 'https://eida.bgs.ac.uk',
    "EARTHSCOPE": "https://service.earthscope.org",
    "EIDA": "http://eida-federator.ethz.ch",
    "EPOSFR": "https://seisdata.epos-france.fr",
    "ETH": "https://eida.ethz.ch",
    "EMSC": "https://www.seismicportal.eu",
    "GEOFON": "https://geofon.gfz.de",
    "GEONET": "https://service.geonet.org.nz",
    "GFZ": "https://geofon.gfz.de",
    "ICGC": "https://ws.icgc.cat",
    "IESDMC": "http://batsws.earth.sinica.edu.tw",
    "IGN": "http://fdsnws.sismologia.ign.es",
    "INGV": "https://webservices.ingv.it",
    "IPGP": "https://ws.ipgp.fr",
    "IRIS": "https://service.earthscope.org",
    "IRISDMC": "https://service.earthscope.org",
    "IRISPH5": "https://service.earthscope.org",
    "ISC": "https://www.isc.ac.uk",
    "KAGSR": "http://sdis.emsd.ru",
    "KNMI": "https://rdsa.knmi.nl",
    "KOERI": "https://eida.koeri.boun.edu.tr",
    "LMU": "https://erde.geophysik.uni-muenchen.de",
    "NCEDC": "https://service.ncedc.org",
    "NIEP": "https://eida-sc3.infp.ro",
    "NOA": "https://eida.gein.noa.gr",
    "NRCAN": "https://earthquakescanada.nrcan.gc.ca",
    "ODC": "https://www.orfeus-eu.org",
    "ORFEUS": "https://www.orfeus-eu.org",
    "RESIF": "https://ws.resif.fr",
    "RASPISHAKE": "https://data.raspberryshake.org",
    "SCEDC": "https://service.scedc.caltech.edu",
    "TEXNET": "http://rtserve.beg.utexas.edu",
    "UIB-NORSAR": "https://eida.geo.uib.no",
    "USGS": "https://earthquake.usgs.gov",
    "USP": "https://sismo.iag.usp.br"
}
URL_MAPPING_SUBPATHS = {
    "IRISPH5": "/ph5ws",
}
URL_DEFAULT_SUBPATH = "/fdsnws"

FDSNWS = ("dataselect", "event", "station")

encoding = sys.getdefaultencoding() or "UTF-8"
platform_ = platform.platform().encode(encoding).decode("ascii", "ignore")
# The default User Agent that will be sent with every request.
DEFAULT_USER_AGENT = "ObsPy/%s (%s, Python %s)" % (
    __version__, platform_, platform.python_version())


# The default parameters. Different services can choose to add more. It always
# contains the long name first and the short name second. If it has no short
# name, it is simply a tuple with only one entry.
DEFAULT_DATASELECT_PARAMETERS = [
    "starttime", "endtime", "network", "station", "location", "channel"]

OPTIONAL_DATASELECT_PARAMETERS = [
    "quality", "minimumlength", "longestonly"]

DEFAULT_STATION_PARAMETERS = [
    "starttime", "endtime", "network", "station", "location", "channel",
    "minlatitude", "maxlatitude", "minlongitude", "maxlongitude", "level"]

OPTIONAL_STATION_PARAMETERS = [
    "startbefore", "startafter", "endbefore", "endafter", "latitude",
    "longitude", "minradius", "maxradius", "includerestricted",
    "includeavailability", "updatedafter", "matchtimeseries", "format"]

DEFAULT_EVENT_PARAMETERS = [
    "starttime", "endtime", "minlatitude", "maxlatitude", "minlongitude",
    "maxlongitude", "mindepth", "maxdepth", "minmagnitude", "maxmagnitude",
    "orderby"]

OPTIONAL_EVENT_PARAMETERS = [
    "latitude", "longitude", "minradius", "maxradius", "magnitudetype",
    "eventtype", "includeallorigins", "includeallmagnitudes",
    "includearrivals", "eventid", "limit", "offset", "catalog", "contributor",
    "updatedafter"]

DEFAULT_PARAMETERS = {
    "dataselect": DEFAULT_DATASELECT_PARAMETERS,
    "event": DEFAULT_EVENT_PARAMETERS,
    "station": DEFAULT_STATION_PARAMETERS}

OPTIONAL_PARAMETERS = {
    "dataselect": OPTIONAL_DATASELECT_PARAMETERS,
    "event": OPTIONAL_EVENT_PARAMETERS,
    "station": OPTIONAL_STATION_PARAMETERS}

PARAMETER_ALIASES = {
    "net": "network",
    "sta": "station",
    "loc": "location",
    "cha": "channel",
    "start": "starttime",
    "end": "endtime",
    "minlat": "minlatitude",
    "maxlat": "maxlatitude",
    "minlon": "minlongitude",
    "maxlon": "maxlongitude",
    "lat": "latitude",
    "lon": "longitude",
    "minmag": "minmagnitude",
    "maxmag": "maxmagnitude",
    "magtype": "magnitudetype",
}


# The default types if none are given. If the parameter can not be found in
# here and has no specified type, the type will be assumed to be a string.
DEFAULT_TYPES = {
    "starttime": UTCDateTime,
    "endtime": UTCDateTime,
    "network": str,
    "station": str,
    "location": str,
    "channel": str,
    "quality": str,
    "minimumlength": float,
    "longestonly": bool,
    "startbefore": UTCDateTime,
    "startafter": UTCDateTime,
    "endbefore": UTCDateTime,
    "endafter": UTCDateTime,
    "maxlongitude": float,
    "minlongitude": float,
    "longitude": float,
    "maxlatitude": float,
    "minlatitude": float,
    "latitude": float,
    "maxdepth": float,
    "mindepth": float,
    "maxmagnitude": float,
    "minmagnitude": float,
    "magnitudetype": str,
    "maxradius": float,
    "minradius": float,
    "level": str,
    "includerestricted": bool,
    "includeavailability": bool,
    "includeallorigins": bool,
    "includeallmagnitudes": bool,
    "includearrivals": bool,
    "matchtimeseries": bool,
    "eventid": str,
    "eventtype": str,
    "limit": int,
    "offset": int,
    "orderby": str,
    "catalog": str,
    "contributor": str,
    "updatedafter": UTCDateTime,
    "format": str}

DEFAULT_VALUES = {
    "starttime": None,
    "endtime": None,
    "network": None,
    "station": None,
    "location": None,
    "channel": None,
    "quality": "B",
    "minimumlength": 0.0,
    "longestonly": False,
    "startbefore": None,
    "startafter": None,
    "endbefore": None,
    "endafter": None,
    "maxlongitude": 180.0,
    "minlongitude": -180.0,
    "longitude": 0.0,
    "maxlatitude": 90.0,
    "minlatitude": -90.0,
    "latitude": 0.0,
    "maxdepth": None,
    "mindepth": None,
    "maxmagnitude": None,
    "minmagnitude": None,
    "magnitudetype": None,
    "maxradius": 180.0,
    "minradius": 0.0,
    "level": "station",
    "includerestricted": True,
    "includeavailability": False,
    "includeallorigins": False,
    "includeallmagnitudes": False,
    "includearrivals": False,
    "matchtimeseries": False,
    "eventid": None,
    "eventtype": None,
    "limit": None,
    "offset": 1,
    "orderby": "time",
    "catalog": None,
    "contributor": None,
    "updatedafter": None,
}

# This creates a services dictionary containing default and optional services,
# with reasonable types and default values, but none are required. Its purpose
# is to look like what would be returned from an actual services query on a
# minimal and very permissive service provider, without actually having to
# do the query.
DEFAULT_SERVICES = {}
for service in ["dataselect", "event", "station"]:
    DEFAULT_SERVICES[service] = {}

    for default_param in DEFAULT_PARAMETERS[service]:
        DEFAULT_SERVICES[service][default_param] = {
            "default_value": DEFAULT_VALUES[default_param],
            "type": DEFAULT_TYPES[default_param],
            "required": False,
        }

    for optional_param in OPTIONAL_PARAMETERS[service]:
        if optional_param == "format":
            if service == "dataselect":
                default_val = "miniseed"
            else:
                default_val = "xml"
        else:
            default_val = DEFAULT_VALUES[optional_param]

        DEFAULT_SERVICES[service][optional_param] = {
            "default_value": default_val,
            "type": DEFAULT_TYPES[optional_param],
            "required": False,
        }

# This list collects WADL parameters that will not be parsed because they are
# not useful for the ObsPy client.
# Current the nodata parameter used by EARTHSCOPE/IRIS is part of that list.
# The ObsPy client relies on the HTTP codes.
WADL_PARAMETERS_NOT_TO_BE_PARSED = ["nodata"]



import numpy as np
import matplotlib.pyplot as plt

import obspy
from obspy.signal.invsim import corn_freq_2_paz


paz_sts2 = {
    'poles': [-0.037004 + 0.037016j, -0.037004 - 0.037016j, -251.33 + 0j,
              - 131.04 - 467.29j, -131.04 + 467.29j],
    'zeros': [0j, 0j],
    'gain': 60077000.0,
    'sensitivity': 2516778400.0}
paz_1hz = corn_freq_2_paz(1.0, damp=0.707)  # 1Hz instrument
paz_1hz['sensitivity'] = 1.0

st = obspy.read()
# make a copy to keep our original data
st_orig = st.copy()

# Simulate instrument given poles, zeros and gain of
# the original and desired instrument
st.simulate(paz_remove=paz_sts2, paz_simulate=paz_1hz)


tr = st[0]
tr_orig = st_orig[0]

t = np.arange(tr.stats.npts) / tr.stats.sampling_rate

plt.subplot(211)
plt.plot(t, tr_orig.data, 'k')
plt.ylabel('STS-2 [counts]')
plt.subplot(212)
plt.plot(t, tr.data, 'k')
plt.ylabel('1Hz Instrument [m/s]')
plt.xlabel('Time [s]')
plt.show()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#  Purpose: Convenience imports for obspy
#   Author: Robert Barsch
#           Moritz Beyreuther
#           Lion Krischer
#           Tobias Megies
#
# Copyright (C) 2008-2014 Robert Barsch, Moritz Beyreuther, Lion Krischer,
#                         Tobias Megies
# -----------------------------------------------------------------------------
"""
ObsPy: A Python Toolbox for seismology/seismological observatories
==================================================================

ObsPy is an open-source project dedicated to provide a Python framework for
processing seismological data. It provides parsers for common file formats
and seismological signal processing routines which allow the manipulation of
seismological time series.

The goal of the ObsPy project is to facilitate rapid application development
for seismology.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import sys
import warnings

# don't change order
# ignore pkg resources deprecation warning for now
# we really only need this here for one single reason: this file here gets
# imported during the pytest startup phase before ignore rules in pytest.ini
# take effect. otherwise we could just add this warning to the ignore list in
# pytest.ini (see #3333)
# The warning capture can be removed after #3333 gets finalized and merged
with warnings.catch_warnings(record=True) as w:
    msg = ('pkg_resources is deprecated as an API')
    warnings.filterwarnings(
        'ignore', message=msg, category=DeprecationWarning, module='obspy')
    from obspy.core.utcdatetime import UTCDateTime  # NOQA
from obspy.core.util import _get_version_string
__version__ = _get_version_string(abbrev=10)
from obspy.core.trace import Trace  # NOQA
from obspy.core.stream import Stream, read
from obspy.core.event import read_events, Catalog
from obspy.core.inventory import read_inventory, Inventory  # NOQA
from obspy.core.util.obspy_types import (  # NOQA
    ObsPyException, ObsPyReadingError)


__all__ = ["UTCDateTime", "Trace", "__version__", "Stream", "read",
           "read_events", "Catalog", "read_inventory", "ObsPyException",
           "ObsPyReadingError"]


# insert supported read/write format plugin lists dynamically in docstrings
from obspy.core.util.base import _add_format_plugin_table


_add_format_plugin_table(read, "waveform", "read", numspaces=4)
_add_format_plugin_table(read_events, "event", "read", numspaces=4)
_add_format_plugin_table(read_inventory, "inventory", "read", numspaces=4)
_add_format_plugin_table(Stream.write, "waveform", "write", numspaces=8)
_add_format_plugin_table(Catalog.write, "event", "write", numspaces=8)
_add_format_plugin_table(Inventory.write, "inventory", "write", numspaces=8)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
The obspy.clients.syngine test suite.
"""
import io
import re
from unittest import mock

import numpy as np
import pytest

import obspy
from obspy.core.util.base import NamedTemporaryFile
from obspy.clients.syngine import Client
from obspy.clients.base import DEFAULT_TESTING_USER_AGENT, ClientHTTPException


BASE_URL = "https://service.earthscope.org/irisws/syngine/1"
pytestmark = pytest.mark.network


class RequestsMockResponse(object):
    def __init__(self):
        self.text = ""
        self.content = b""
        self.status_code = 200
        self._json = {}

    def json(self):
        return self._json


class TestClient():
    """
    Test cases for obspy.clients.iris.client.Client.
    """
    c = Client(user_agent=DEFAULT_TESTING_USER_AGENT)

    def test_get_model_info_mock(self):
        """
        Mock test for the get_model_info() method.
        """
        with mock.patch("requests.get") as p:
            r = RequestsMockResponse()
            r._json["slip"] = [0.0, 1.0, 2.0]
            r._json["sliprate"] = [0.0, 1.0, 2.0]
            p.return_value = r
            self.c.get_model_info("test_model")

        assert p.call_count == 1
        assert p.call_args[1]["url"] == \
            'https://service.earthscope.org/irisws/syngine/1/info'
        assert p.call_args[1]["params"] == {'model': 'test_model'}
        assert p.call_args[1]["headers"] == \
            {'User-Agent': DEFAULT_TESTING_USER_AGENT}

    def test_get_model_info(self):
        """
        Actual test for the get_model_info() method.
        """
        info = self.c.get_model_info("test")

        assert isinstance(info, obspy.core.AttribDict)
        # Check two random keys.
        assert info.dump_type == "displ_only"
        assert info.time_scheme == "newmark2"
        # Check that both arrays have been converted to numpy arrays.
        assert isinstance(info.slip, np.ndarray)
        assert isinstance(info.sliprate, np.ndarray)

    def test_get_available_models_mock(self):
        with mock.patch("requests.get") as p:
            p.return_value = RequestsMockResponse()
            self.c.get_available_models()

        assert p.call_count == 1
        assert p.call_args[1]["url"] == \
            'https://service.earthscope.org/irisws/syngine/1/models'
        assert p.call_args[1]["params"] is None
        assert p.call_args[1]["headers"] == \
            {'User-Agent': DEFAULT_TESTING_USER_AGENT}

    def test_get_available_models(self):
        models = self.c.get_available_models()
        assert isinstance(models, dict)
        assert len(models) > 3
        assert "ak135f_5s" in models
        # Check random key.
        assert models["ak135f_5s"]["components"] == "vertical and horizontal"

    def test_get_service_version_mock(self):
        with mock.patch("requests.get") as p:
            p.return_value = RequestsMockResponse()
            p.return_value.text = "1.2.3"
            version = self.c.get_service_version()

        assert version == "1.2.3"

        assert p.call_count == 1
        assert p.call_args[1]["url"] == \
            'https://service.earthscope.org/irisws/syngine/1/version'
        assert p.call_args[1]["params"] is None
        assert p.call_args[1]["headers"] == \
            {'User-Agent': DEFAULT_TESTING_USER_AGENT}

    def test_get_waveforms_mock(self):
        """
        Test the queries from the IRIS syngine website and see if they
        produce the correct URLS.
        """
        r = RequestsMockResponse()
        with io.BytesIO() as buf:
            obspy.read()[0].write(buf, format="mseed")
            buf.seek(0, 0)
            r.content = buf.read()

        # https://service.earthscope.org/irisws/syngine/1/query?network=IU&
        # station=ANMO&components=ZRT&eventid=GCMT:M110302J
        with mock.patch("requests.get") as p:
            p.return_value = r
            st = self.c.get_waveforms(model="ak135f_5s",
                                      network="IU", station="ANMO",
                                      components="ZRT",
                                      eventid="GCMT:M110302J")

        assert isinstance(st, obspy.Stream)

        assert p.call_count == 1
        assert p.call_args[1]["url"] == \
            "https://service.earthscope.org/irisws/syngine/1/query"
        assert p.call_args[1]["params"] == {
            "components": "ZRT",
            "eventid": "GCMT:M110302J",
            "format": "miniseed",
            "model": "ak135f_5s",
            "network": "IU",
            "station": "ANMO"}
        assert p.call_args[1]["headers"] == \
            {"User-Agent": DEFAULT_TESTING_USER_AGENT}

        # https://service.earthscope.org/irisws/syngine/1/query?network=_GSN&
        # components=Z&eventid=GCMT:M110302J&endtime=1800
        with mock.patch("requests.get") as p:
            p.return_value = r
            st = self.c.get_waveforms(model="ak135f_5s",
                                      network="_GSN",
                                      components="Z",
                                      endtime=1800.0,
                                      eventid="GCMT:M110302J")

        assert isinstance(st, obspy.Stream)

        assert p.call_count == 1
        assert p.call_args[1]["url"] == \
            "https://service.earthscope.org/irisws/syngine/1/query"
        assert p.call_args[1]["params"] == {
            "components": "Z",
            "endtime": 1800.0,
            "eventid": "GCMT:M110302J",
            "format": "miniseed",
            "model": "ak135f_5s",
            "network": "_GSN"}
        assert p.call_args[1]["headers"] == \
            {"User-Agent": DEFAULT_TESTING_USER_AGENT}

        # https://service.earthscope.org/irisws/syngine/1/query?network=_GSN&
        # components=Z&eventid=GCMT:M110302J&starttime=P-10&endtime=ScS%2B60
        with mock.patch("requests.get") as p:
            p.return_value = r
            st = self.c.get_waveforms(model="ak135f_5s",
                                      network="_GSN",
                                      components="Z",
                                      starttime="P-10",
                                      endtime="ScS+60",
                                      eventid="GCMT:M110302J")

        assert isinstance(st, obspy.Stream)

        assert p.call_count == 1
        assert p.call_args[1]["url"] == \
            "https://service.earthscope.org/irisws/syngine/1/query"
        assert p.call_args[1]["params"] == {
            "components": "Z",
            "starttime": "P-10",
            "endtime": "ScS+60",
            "eventid": "GCMT:M110302J",
            "format": "miniseed",
            "model": "ak135f_5s",
            "network": "_GSN"}
        assert p.call_args[1]["headers"] == \
            {"User-Agent": DEFAULT_TESTING_USER_AGENT}

    def test_error_handling_arguments(self):
        # Floating points value
        with pytest.raises(ValueError):
            self.c.get_waveforms(model="test", receiverlatitude="a")
        # Int.
        with pytest.raises(ValueError):
            self.c.get_waveforms(model="test", kernelwidth="a")
        # Time.
        with pytest.raises(TypeError):
            self.c.get_waveforms(model="test", origintime="a")

    def test_source_mechanisms_mock(self):
        r = RequestsMockResponse()
        with io.BytesIO() as buf:
            obspy.read()[0].write(buf, format="mseed")
            buf.seek(0, 0)
            r.content = buf.read()

        with mock.patch("requests.get") as p:
            p.return_value = r
            self.c.get_waveforms(model="ak135f_5s",
                                 sourcemomenttensor=[1, 2, 3, 4, 5, 6])
        assert p.call_count == 1
        assert p.call_args[1]["url"] == \
            "https://service.earthscope.org/irisws/syngine/1/query"
        assert p.call_args[1]["params"] == {
            "model": "ak135f_5s",
            "format": "miniseed",
            "sourcemomenttensor": "1,2,3,4,5,6"}

        with mock.patch("requests.get") as p:
            p.return_value = r
            self.c.get_waveforms(model="ak135f_5s",
                                 sourcedoublecouple=[1, 2, 3, 4])
        assert p.call_count == 1
        assert p.call_args[1]["url"] == \
            "https://service.earthscope.org/irisws/syngine/1/query"
        assert p.call_args[1]["params"] == {
            "model": "ak135f_5s",
            "format": "miniseed",
            "sourcedoublecouple": "1,2,3,4"}

        with mock.patch("requests.get") as p:
            p.return_value = r
            self.c.get_waveforms(model="ak135f_5s",
                                 sourceforce=[3.32, 4.23, 5.11])
        assert p.call_count == 1
        assert p.call_args[1]["url"] == \
            "https://service.earthscope.org/irisws/syngine/1/query"
        assert p.call_args[1]["params"] == {
            "model": "ak135f_5s",
            "format": "miniseed",
            "sourceforce": "3.32,4.23,5.11"}

    def test_error_handling(self):
        """
        Tests the error handling. The clients just pass on most things to
        syngine and rely on the service for the error detection.
        """
        # Wrong components.
        msg = re.compile("HTTP code 400 when.*Unrecognized component",
                         re.DOTALL)
        with pytest.raises(ClientHTTPException, match=msg):
            self.c.get_waveforms(
                model="ak135f_5s", eventid="GCMT:C201002270634A",
                station="ANMO", network="IU", components="ABC")

    def test_bulk_waveform_download_mock(self):
        """
        Mock the bulk download requests to test the payload generation.
        """
        r = RequestsMockResponse()
        with io.BytesIO() as buf:
            obspy.read()[0].write(buf, format="mseed")
            buf.seek(0, 0)
            r.content = buf.read()

        payload = []

        def side_effect(*args, **kwargs):
            payload[:] = [kwargs["data"].decode()]
            return r

        # Test simple lists first.
        with mock.patch("requests.post") as p:
            p.side_effect = side_effect
            self.c.get_waveforms_bulk(
                model="ak135f_5s", bulk=[
                    [1.0, 2.0],
                    (2.0, 3.0),
                    ("AA", "BB")])

        assert payload[0] == "\n".join([
            "model=ak135f_5s",
            "format=miniseed",
            "1.0 2.0",
            "2.0 3.0",
            "AA BB\n"])

        # A couple more parameters
        with mock.patch("requests.post") as p:
            p.side_effect = side_effect
            self.c.get_waveforms_bulk(
                model="ak135f_5s", bulk=[
                    [1.0, 2.0],
                    (2.0, 3.0),
                    ("AA", "BB")],
                format="miniseed",
                sourcemomenttensor=[1, 2, 3, 4, 5, 6])

        assert payload[0] == "\n".join([
            "model=ak135f_5s",
            "format=miniseed",
            "sourcemomenttensor=1,2,3,4,5,6",
            "1.0 2.0",
            "2.0 3.0",
            "AA BB\n"])

        # A couple of dictionaries.
        with mock.patch("requests.post") as p:
            p.side_effect = side_effect
            self.c.get_waveforms_bulk(
                model="ak135f_5s", bulk=[
                    {"network": "IU", "station": "ANMO"},
                    {"latitude": 12, "longitude": 13.1},
                    {"latitude": 12, "longitude": 13.1, "networkcode": "IU"},
                    {"latitude": 12, "longitude": 13.1, "stationcode": "ANMO"},
                    {"latitude": 12, "longitude": 13.1, "locationcode": "00"},
                    {"latitude": 12, "longitude": 13.1, "networkcode": "IU",
                     "stationcode": "ANMO", "locationcode": "00"}],
                format="miniseed", eventid="GCMT:C201002270634A")

        assert payload[0] == "\n".join([
            "model=ak135f_5s",
            "eventid=GCMT:C201002270634A",
            "format=miniseed",
            "IU ANMO",
            "12 13.1",
            "12 13.1 NETCODE=IU",
            "12 13.1 STACODE=ANMO",
            "12 13.1 LOCCODE=00",
            "12 13.1 NETCODE=IU STACODE=ANMO LOCCODE=00\n"])

    def test_get_waveforms(self):
        """
        Test get_waveforms() and get_waveforms_bulk() by actually downloading
        some things.

        Use the 'test' model which does not produce useful seismograms but
        is quick to test.
        """
        st = self.c.get_waveforms(model="test", network="IU", station="ANMO",
                                  eventid="GCMT:C201002270634A",
                                  components="Z")
        assert len(st) == 1
        # Download exactly the same with a bulk request and check the result
        # is the same!
        st_bulk = self.c.get_waveforms_bulk(
            model="test", bulk=[("IU", "ANMO")],
            eventid="GCMT:C201002270634A", components="Z")
        assert len(st_bulk) == 1
        assert st == st_bulk

        # Test phase relative times. This tests that everything is correctly
        # encoded and what not.
        st = self.c.get_waveforms(model="test", network="IU", station="ANMO",
                                  eventid="GCMT:C201002270634A",
                                  starttime="P-10", endtime="P+20",
                                  components="Z")
        assert len(st) == 1
        st_bulk = self.c.get_waveforms_bulk(
            model="test", bulk=[("IU", "ANMO")],
            starttime="P-10", endtime="P+20",
            eventid="GCMT:C201002270634A", components="Z")
        assert len(st_bulk) == 1
        assert st == st_bulk

        # One to test a source mechanism
        st = self.c.get_waveforms(model="test", network="IU", station="ANMO",
                                  sourcemomenttensor=[1, 2, 3, 4, 5, 6],
                                  sourcelatitude=10, sourcelongitude=20,
                                  sourcedepthinmeters=100,
                                  components="Z")
        assert len(st) == 1
        st_bulk = self.c.get_waveforms_bulk(
            model="test", bulk=[("IU", "ANMO")],
            sourcemomenttensor=[1, 2, 3, 4, 5, 6],
            sourcelatitude=10, sourcelongitude=20,
            sourcedepthinmeters=100,
            components="Z")
        assert len(st_bulk) == 1
        assert st == st_bulk

        # One more to test actual time values.
        st = self.c.get_waveforms(
            model="test", network="IU", station="ANMO",
            origintime=obspy.UTCDateTime(2015, 1, 2, 3, 0, 5),
            starttime=obspy.UTCDateTime(2015, 1, 2, 3, 4, 5),
            endtime=obspy.UTCDateTime(2015, 1, 2, 3, 10, 5),
            sourcemomenttensor=[1, 2, 3, 4, 5, 6],
            sourcelatitude=10, sourcelongitude=20,
            sourcedepthinmeters=100,
            components="Z")
        assert len(st) == 1
        st_bulk = self.c.get_waveforms_bulk(
            model="test", bulk=[("IU", "ANMO")],
            origintime=obspy.UTCDateTime(2015, 1, 2, 3, 0, 5),
            starttime=obspy.UTCDateTime(2015, 1, 2, 3, 4, 5),
            endtime=obspy.UTCDateTime(2015, 1, 2, 3, 10, 5),
            sourcemomenttensor=[1, 2, 3, 4, 5, 6],
            sourcelatitude=10, sourcelongitude=20,
            sourcedepthinmeters=100,
            components="Z")
        assert len(st_bulk) == 1
        assert st == st_bulk

    def test_saving_directly_to_file(self):
        # Save to a filename.
        with NamedTemporaryFile() as tf:
            filename = tf.name
            st = self.c.get_waveforms(
                model="test", network="IU", station="ANMO",
                eventid="GCMT:C201002270634A", starttime="P-10",
                endtime="P+10", components="Z", filename=tf)
            # No return value.
            assert st is None

            st = obspy.read(filename)
            assert len(st) == 1

        # Save to an open file-like object.
        with io.BytesIO() as buf:
            st = self.c.get_waveforms(
                model="test", network="IU", station="ANMO",
                eventid="GCMT:C201002270634A", starttime="P-10",
                endtime="P+10", components="Z", filename=buf)
            # No return value.
            assert st is None

            buf.seek(0, 0)
            st = obspy.read(buf)
            assert len(st) == 1

    def test_reading_saczip_files(self):
        with pytest.warns(UserWarning,
                          match="Sample spacing read from SAC"):
            st = self.c.get_waveforms(
                model="test", network="IU", station="ANMO",
                eventid="GCMT:C201002270634A", starttime="P-10",
                endtime="P+10", components="Z", format="saczip")
        assert len(st) == 1
        # Same with bulk request.
        with pytest.warns(UserWarning,
                          match="Sample spacing read from SAC"):
            st_bulk = self.c.get_waveforms_bulk(
                model="test", bulk=[("IU", "ANMO")],
                eventid="GCMT:C201002270634A", starttime="P-10",
                endtime="P+10", components="Z", format="saczip")
        assert len(st_bulk) == 1

        assert st == st_bulk

    def test_bulk_waveform_send_custom_payload(self):
        """
        The get_waveforms_bulk() method can send a custom payload.
        """
        r = RequestsMockResponse()
        with io.BytesIO() as buf:
            obspy.read()[0].write(buf, format="mseed")
            buf.seek(0, 0)
            r.content = buf.read()

        payload = []

        def side_effect(*args, **kwargs):
            payload[:] = [kwargs["data"]]
            return r

        # Test simple lists first.
        with mock.patch("requests.post") as p:
            p.side_effect = side_effect
            self.c.get_waveforms_bulk(
                model="ak135f_5s", bulk=[], data=b"1234\n5678")

        assert payload[0] == b"1234\n5678"



# -*- coding: utf-8 -*-
import io

import numpy as np

from obspy import read
from obspy.io.nied.knet import _is_knet_ascii


class TestKnetReading():
    """
    Test reading of K-NET and KiK-net ASCII format files from a file.
    """
    def test_read_knet_ascii(self, testdata):
        testfile = testdata['test.knet']
        tr = read(testfile)[0]
        tr.data *= tr.stats.calib
        tr.data -= tr.data.mean()
        max = np.abs(tr.data).max() * 100  # Maximum acc converted to gal
        np.testing.assert_array_almost_equal(max, tr.stats.knet.accmax,
                                             decimal=3)
        duration = int(tr.stats.endtime - tr.stats.starttime + 0.5)
        assert duration == int(tr.stats.knet.duration)

    def test_read_knet_ascii_from_open_files(self, testdata):
        """
        Test reading of K-NET and KiK-net ASCII format files from an open file.
        """
        testfile = testdata['test.knet']
        tr = read(testfile)[0]
        tr.data *= tr.stats.calib
        tr.data -= tr.data.mean()
        max = np.abs(tr.data).max() * 100  # Maximum acc converted to gal
        np.testing.assert_array_almost_equal(max, tr.stats.knet.accmax,
                                             decimal=3)
        duration = int(tr.stats.endtime - tr.stats.starttime + 0.5)
        assert duration == int(tr.stats.knet.duration)

    def test_read_knet_ascii_from_bytes_io(self, testdata):
        """
        Tests that reading of K-NET and KiK-net ASCII format files from a
        BytesIO object works.
        """
        testfile = testdata['test.knet']
        with open(testfile, "rb") as fh:
            buf = io.BytesIO(fh.read())

        with buf:
            tr = read(buf)[0]
            tr.data *= tr.stats.calib
            tr.data -= tr.data.mean()
            max = np.abs(tr.data).max() * 100  # Maximum acc converted to gal
            np.testing.assert_array_almost_equal(max, tr.stats.knet.accmax,
                                                 decimal=3)
            duration = int(tr.stats.endtime - tr.stats.starttime + 0.5)
            assert duration == int(tr.stats.knet.duration)

    def test_station_name_hack(self, testdata):
        """
        Station names in K-NET and KiK-net are 6 characters long which does not
        conform with the SEED standard. Test hack to write the last 2
        characters of the station name into the location field.
        """
        testfile = testdata['test.knet']
        tr = read(testfile, convert_stnm=True)[0]
        assert tr.stats.location == '13'

    def test_is_knet_ascii(self, testdata, datapath):
        """
        This tests the _is_knet_ascii method by just validating that each file
        in the data directory is a K-NET ascii file and each file in the
        working directory is not.

        The filenames are hard coded so the test will not fail with future
        changes in the structure of the package.
        """
        # K-NET file names.
        knet_filenames = ['test.knet']

        # Non K-NET file names.
        non_knet_filenames = ['test_knet_reading.py',
                              '__init__.py']
        # Loop over K-NET files
        for _i in knet_filenames:
            filename = testdata[_i]
            is_knet = _is_knet_ascii(filename)
            assert is_knet
        # Loop over non K-NET files
        for _i in non_knet_filenames:
            filename = datapath.parent / _i
            is_knet = _is_knet_ascii(filename)
            assert not is_knet



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test suite for the arclink inventory reader.

Modified after obspy.io.stationXML
    > obspy.obspy.io.stationxml.core.py

:author:
    Mathijs Koymans (koymans@knmi.nl), 11.2015 - [Jollyfant@GitHub]

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from obspy.core.inventory import read_inventory
from obspy.io.arclink.inventory import validate_arclink_xml, SCHEMA_NAMESPACE
import pytest


class TestArclinkInventory():

    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        """
        Read example stationXML and arclink format to Inventory
        """
        self.arclink_xml_path = testdata["arclink_inventory.xml"]
        self.station_xml_path = testdata["gols_station.xml"]
        self.arclink_xml_poly = testdata["arclink_inventory_poly.xml"]
        self.arclink_afc_path = testdata["arclink_afc.xml"]
        self.station_afc_path = testdata["station_afc.xml"]

    def test_publicid_slash(self, testdata):
        v = read_inventory(testdata["public-id-slash.xml"])
        assert len(v[0][0][0].response.response_stages) == 11

    def test_analogue_filter_chain(self):
        """
        Test analogue filter chain and compare to stationXML equivalent
        """
        arclink_inv = read_inventory(self.arclink_afc_path)
        station_inv = read_inventory(self.station_afc_path)

        arclink_resp = arclink_inv[0][0][0].response.response_stages
        station_resp = station_inv[0][0][0].response.response_stages

        for arclink_stage, station_stage in zip(arclink_resp, station_resp):

            assert arclink_stage.stage_gain == \
                             station_stage.stage_gain
            assert arclink_stage.stage_sequence_number == \
                station_stage.stage_sequence_number

            for arc_cs, sta_cs in zip(arclink_stage.__dict__.items(),
                                      station_stage.__dict__.items()):

                if arc_cs[0] in ['name', 'resource_id2']:
                    continue

                assert arc_cs == sta_cs

    def test_validate_inventories_against_schema(self):
        assert validate_arclink_xml(self.arclink_xml_path)[0]
        assert not validate_arclink_xml(self.station_xml_path)[0]

    def test_raise_polynomial(self):
        msg = ("responsePolynomial not implemented. Contact the ObsPy "
               "developers")
        with pytest.raises(NotImplementedError, match=msg):
            read_inventory(self.arclink_xml_poly)

    @pytest.mark.filterwarnings("ignore:Attribute 'storage_format'.*removed")
    def test_auto_read_arclink_xml(self):
        arclink_inv = read_inventory(self.arclink_xml_path)
        assert arclink_inv is not None

        station_inv = read_inventory(self.station_xml_path)

        arclink_inv = arclink_inv.select(station="GOLS")
        channels = ['HH1', 'HH2', 'HHZ', 'BHZ', 'BH1', 'BH2', 'LH1', 'LH2',
                    'LHZ']

        for channel in channels:
            arclink_cha = arclink_inv.select(channel=channel)
            station_cha = station_inv.select(channel=channel)
            for arc, st_xml in zip(arclink_cha[0][0], station_cha[0][0]):
                assert arc.code == st_xml.code
                assert arc.latitude == st_xml.latitude
                assert arc.longitude == st_xml.longitude
                assert arc.depth == st_xml.depth
                assert arc.azimuth == st_xml.azimuth
                assert arc.dip == st_xml.dip
                assert arc.sample_rate == st_xml.sample_rate
                assert arc.start_date == st_xml.start_date
                assert arc.end_date == st_xml.end_date
                # reading stationxml will ignore old StationXML 1.0 defined
                # StorageFormat, Arclink Inventory XML and SC3ML get it stored
                # in extra now
                assert st_xml.storage_format is None
                assert arc.storage_format is None
                assert arc.extra['format'] == \
                    {'namespace': SCHEMA_NAMESPACE, 'value': None}

                cdisps = "clock_drift_in_seconds_per_sample"
                assert getattr(arc, cdisps) == getattr(st_xml, cdisps)

                # Compare datalogger element
                for arc_el, st_xml_el in zip(arc.data_logger.__dict__.items(),
                                             st_xml.data_logger.__dict__
                                             .items()):
                    assert arc_el == st_xml_el

                # Compare sensor element
                for arc_sen, st_xml_sen in zip(arc.sensor.__dict__.items(),
                                               st_xml.sensor.__dict__.items()):
                    # Skip the type; set for stationXML not for ArclinkXML
                    if arc_sen[0] == 'type':
                        continue
                    assert arc_sen == st_xml_sen

                # Same number of response stages
                assert len(arc.response.response_stages) == \
                    len(st_xml.response.response_stages)

                # Check the response stages
                for arc_resp, sta_resp in zip(arc.response.response_stages,
                                              st_xml.response.response_stages):

                    assert arc_resp.stage_gain == sta_resp.stage_gain
                    assert arc_resp.stage_sequence_number == \
                        sta_resp.stage_sequence_number

                    for arc_cs, sta_cs in zip(arc_resp.__dict__.items(),
                                              sta_resp.__dict__.items()):

                        # Name can be different and resource_id is not given
                        # in stationXML
                        if arc_cs[0] in ['name', 'resource_id2']:
                            continue

                        # Floating point precision troubles..
                        # 0.021099999999999997 != 0.0211
                        if isinstance(arc_cs[1], float):
                            assert round(abs(arc_cs[1]-sta_cs[1]), 7) == 0
                        else:
                            assert arc_cs == sta_cs



import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.pyplot as plt

from obspy import read_inventory, read_events

# Set up a custom projection
projection = ccrs.AlbersEqualArea(
    central_longitude=35,
    central_latitude=50,
    standard_parallels=(40, 42)
)

# Set up a figure
fig = plt.figure(dpi=150)
ax = fig.add_subplot(111, projection=projection)
ax.set_extent((-15., 75., 15., 80.))

# Draw standard features
ax.gridlines()
ax.coastlines()
ax.stock_img()
ax.add_feature(cfeature.BORDERS)

ax.set_title("Albers Equal Area Projection")

# Now, let's plot some data on the map
inv = read_inventory()
inv.plot(fig=fig, show=False)
cat = read_events()
cat.plot(fig=fig, show=False, title="", colorbar=False)

plt.show()



# -*- coding: utf-8 -*-
"""
SeedLink request client for ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import fnmatch
import warnings

import numpy as np
from lxml import etree

from obspy import Stream
from .slclient import SLClient, SLPacket
from .client.seedlinkconnection import SeedLinkConnection
from obspy.io.mseed.core import _read_mseed


class Client(object):
    """
    SeedLink request client.

    This client is intended for requests of specific, finite time windows.
    To work with continuous realtime data streams please see
    :class:`~obspy.clients.seedlink.slclient.SLClient` and
    :class:`~obspy.clients.seedlink.easyseedlink.EasySeedLinkClient`.

    :type server: str
    :param server: Server name or IP address to connect to (e.g.
        "localhost", "rtserver.ipgp.fr")
    :type port: int
    :param port: Port at which the seedlink server is operating (default is
        `18000`).
    :type timeout: float
    :param timeout: Network timeout for low-level network connection in
        seconds.
    :type debug: bool
    :param debug: Switches on debugging output.
    """
    def __init__(self, server, port=18000, timeout=20, debug=False):
        """
        Initializes the SeedLink request client.
        """
        self.timeout = timeout
        self.debug = debug
        self._server_url = "%s:%i" % (server, port)
        self._station_cache = None
        self._station_cache_level = None

    def _init_client(self):
        """
        Make fresh connection to seedlink server

        Should be done before any request to server, since SLClient keeps
        things like multiselect etc for subsequent requests
        """
        self._slclient = SLClient(timeout=self.timeout)

    def _connect(self):
        """
        Open new connection to seedlink server.
        """
        self._slclient.slconn = SeedLinkConnection(timeout=self.timeout)
        self._slclient.slconn.set_sl_address(self._server_url)
        self._slclient.slconn.netto = self.timeout

    def get_waveforms(self, network, station, location, channel, starttime,
                      endtime):
        """
        Request waveform data from the seedlink server.

        >>> from obspy import UTCDateTime
        >>> client = Client('rtserver.ipgp.fr')
        >>> t = UTCDateTime() - 1500
        >>> st = client.get_waveforms("G", "FDFM", "10", "BHZ", t, t + 5)
        >>> print(st)  # doctest: +ELLIPSIS
        1 Trace(s) in Stream:
        G.FDFM.10.BHZ | 20... | 20.0 Hz, ... samples

        Most servers support '?' single-character wildcard in location and
        channel code fields:

        >>> st = client.get_waveforms("G", "FDFM", "1?", "B??", t, t + 5)
        >>> st = st.sort(reverse=True)
        >>> print(st)  # doctest: +ELLIPSIS
        3 Trace(s) in Stream:
        G.FDFM.10.BHZ | 20... | 20.0 Hz, ... samples
        G.FDFM.10.BHN | 20... | 20.0 Hz, ... samples
        G.FDFM.10.BHE | 20... | 20.0 Hz, ... samples

        Depending on server capabilities, '*' multi-character wildcards might
        work in any parameter:

        >>> st = client.get_waveforms("*", "FDFM", "1*", "B*", t, t + 5)
        >>> st = st.sort(reverse=True)
        >>> print(st)  # doctest: +ELLIPSIS
        3 Trace(s) in Stream:
        G.FDFM.10.BHZ | 20... | 20.0 Hz, ... samples
        G.FDFM.10.BHN | 20... | 20.0 Hz, ... samples
        G.FDFM.10.BHE | 20... | 20.0 Hz, ... samples

        .. note::

            Support of wildcards strongly depends on the queried seedlink
            server. In general, '?' as single character wildcard seems to work
            well in location code and channel code fields for most servers.
            Usage of '*' relies on the server supporting info requests on
            station or even channel level, see :meth:`Client.get_info()`.

        :type network: str
        :param network: Network code. See note on wildcards above.
        :type station: str
        :param station: Station code. See note on wildcards above.
        :type location: str
        :param location: Location code. See note on wildcards above.
        :type channel: str
        :param channel: Channel code. See note on wildcards above.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start time of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End time of requested time window.
        """
        # need to do an info request?
        if any('*' in x for x in (network, station, location, channel)) \
                or any('?' in x for x in (network, station)):
            # need to do an info request on channel level?
            if any('*' in x for x in (location, channel)):
                info = self.get_info(network=network, station=station,
                                     location=location, channel=channel,
                                     level='channel', cache=True)
                multiselect = ["%s_%s:%s%s" % (net, sta, loc, cha)
                               for net, sta, loc, cha in info]
            # otherwise keep location/channel wildcards and do request on
            # station level only
            else:
                info = self.get_info(network=network, station=station,
                                     level='station', cache=True)
                multiselect = ["%s_%s:%s%s" % (net, sta, location, channel)
                               for net, sta in info]
            multiselect = ','.join(multiselect)
            return self._multiselect_request(multiselect, starttime, endtime)

        # if no info request is needed, we just work with the given input
        # (might have some '?' wildcards in loc/cha)
        if len(location) > 2:
            msg = ("Location code ('%s') only supports a maximum of 2 "
                   "characters.") % location
            raise ValueError(msg)
        elif len(location) == 1:
            msg = ("Single character location codes that are not an '*' are "
                   "untested.")
            warnings.warn(msg)
        if location:
            loccha = "%2s%3s" % (location, channel)
        else:
            loccha = channel
        seedlink_id = "%s_%s:%s" % (network, station, loccha)
        return self._multiselect_request(seedlink_id, starttime, endtime)

    def _multiselect_request(self, multiselect, starttime, endtime):
        """
        Make a multiselect request to underlying seedlink client

        Multiselect string is one or more comma separated
        network/station/location/channel combinations as defined by seedlink
        standard, e.g.
        "NETWORK_STATION:LOCATIONCHANNEL,NETWORK_STATION:LOCATIONCHANNEL"
        where location+channel may contain '?' characters but should be exactly
        5 characters long.

        :rtype: :class:`~obspy.core.stream.Stream`
        """
        self._init_client()
        self._slclient.multiselect = multiselect
        self._slclient.begin_time = starttime
        self._slclient.end_time = endtime
        self._connect()
        self._slclient.initialize()
        self.stream = Stream()

        # holds all data received by the client
        self.bulk_data_container = list()

        self._slclient.run(packet_handler=self._fast_packet_handler)

        # process the data all in one go, much more efficient
        input = np.concatenate(self.bulk_data_container)
        self.bulk_data_container = list()
        stream = _read_mseed(input)

        stream.trim(starttime, endtime)
        self.stream = None
        return stream

    def get_info(self, network=None, station=None, location=None, channel=None,
                 level='station', cache=True, warn_on_excluded_stations=False):
        """
        Request available stations information from the seedlink server.

        Supports ``fnmatch`` wildcards, e.g. ``*`` and ``?``, in ``network``,
        ``station``, ``location`` and ``channel``.

        >>> client = Client('rtserver.ipgp.fr')
        >>> info = client.get_info(station="FDFM")
        >>> print(info)
        [('G', 'FDFM')]
        >>> info = client.get_info(
        ...     station="FD?M", channel='[LBH]HZ', location='10',
        ...     level='channel')
        >>> print(info)  # doctest: +NORMALIZE_WHITESPACE
        [('G', 'FDFM', '10', 'BHZ'), ('G', 'FDFM', '10', 'HHZ'),
         ('G', 'FDFM', '10', 'LHZ')]

        Available station information is cached after the first request to the
        server, so use ``cache=False`` on subsequent requests if there is a
        need to force fetching new information from the server (should only
        concern programs running in background for a very long time).

        .. note::
            Stations/channels are excluded from the results for which the
            server indicates it is serving them in general but it also states
            no data are in ring buffer currently.
            If interested in these "no data" stations/channels, either set
            ``warn_on_excluded_stations=True`` which will show a warning
            message with excluded stations or use ``debug=True`` when
            initializing the client which will print the raw server ``seedlink
            INFO`` xml response which will show these stations listed with
            ``begin_seq`` and ``end_seq`` both with value ``'000000'``.

        :type network: str
        :param network: Network code. Supports ``fnmatch`` wildcards, e.g.
            ``*`` and ``?``.
        :type station: str
        :param station: Station code. Supports ``fnmatch`` wildcards, e.g.
            ``*`` and ``?``.
        :type location: str
        :param location: Location code. Supports ``fnmatch`` wildcards, e.g.
            ``*`` and ``?``.
        :type channel: str
        :param channel: Channel code. Supports ``fnmatch`` wildcards, e.g.
            ``*`` and ``?``.
        :type cache: bool
        :param cache: Subsequent function calls are cached, use ``cache=False``
            to force fetching station metadata again from the server.
        :type warn_on_excluded_stations: bool
        :param warn_on_excluded_stations: Whether to show a warning for
            stations that are excluded from the results because the server
            indicates there is no data currently available.
        :rtype: list
        :returns: list of 2-tuples (or 4-tuples with ``level='channel'``) with
            network/station (network/station/location/channel, respectively)
            code combinations for which data is served by the server.
        """
        if level not in ('station', 'channel'):
            msg = "Invalid option for 'level': '%s'" % str(level)
            raise ValueError(msg)
        if level == 'station' and \
                any(x is not None for x in (location, channel)):
            msg = ("location and channel options are ignored in get_info() if "
                   "level='station'.")
            warnings.warn(msg)
        # deteremine if we have a usable cache and check if it is at least the
        # requested level of detail
        if cache and self._station_cache is not None \
                and level in ('station', self._station_cache_level):
            if level == 'station':
                if self._station_cache_level == 'station':
                    info = [(net, sta) for net, sta in self._station_cache
                            if fnmatch.fnmatch(net, network or '*') and
                            fnmatch.fnmatch(sta, station or '*')]
                    return sorted(info)
                else:
                    info = [(net, sta) for net, sta, loc, cha
                            in self._station_cache
                            if fnmatch.fnmatch(net, network or '*') and
                            fnmatch.fnmatch(sta, station or '*')]
                    return sorted(set(info))
            info = [(net, sta, loc, cha) for net, sta, loc, cha in
                    self._station_cache if
                    fnmatch.fnmatch(net, network or '*') and
                    fnmatch.fnmatch(sta, station or '*') and
                    fnmatch.fnmatch(loc, location or '*') and
                    fnmatch.fnmatch(cha, channel or '*')]
            return sorted(info)

        self._init_client()
        if level == 'station':
            self._slclient.infolevel = "STATIONS"
        elif level == 'channel':
            self._slclient.infolevel = "STREAMS"
        self._slclient.verbose = 1
        self._connect()
        self._slclient.initialize()
        # self._slclient.run()
        self._slclient.run(packet_handler=self._packet_handler)
        info = self._slclient.slconn.info_string
        try:
            xml = etree.fromstring(info)
        except ValueError as e:
            msg = 'Unicode strings with encoding declaration are not supported'
            if msg not in str(e):
                raise
            parser = etree.XMLParser(encoding='utf-8')
            xml = etree.fromstring(info.encode('utf-8'), parser=parser)
        station_cache = set()
        excluded_stations = set()
        for tag in xml.xpath('./station'):
            net = tag.attrib['network']
            sta = tag.attrib['name']
            item = (net, sta)
            if level == 'channel':
                subtags = tag.xpath('./stream')
                # If no data is in ring buffer (e.g. station outage?) then it
                # seems the seedlink server replies with no subtags for the
                # channels
                if not subtags:
                    excluded_stations.add(item)
                    continue
                for subtag in subtags:
                    loc = subtag.attrib['location']
                    cha = subtag.attrib['seedname']
                    station_cache.add(item + (loc, cha))
            elif level == 'station':
                # remove stations that seem to have no data
                if all(tag.attrib[key] == '000000'
                       for key in ('begin_seq', 'end_seq')):
                    excluded_stations.add(item)
                    continue
                station_cache.add(item)
            else:
                raise NotImplementedError()
        # change results to an Inventory object
        self._station_cache = station_cache
        self._station_cache_level = level
        if warn_on_excluded_stations and excluded_stations:
            msg = ('Some stations were excluded from results because server '
                   'indicates no data available (use debug=True in Client '
                   'initialization for details, suppress warning with '
                   'warn_on_excluded_stations=False): ')
            msg += ', '.join('.'.join(item)
                             for item in sorted(excluded_stations))
            warnings.warn(msg)
        return self.get_info(
            network=network, station=station, location=location,
            channel=channel, cache=True, level=level)

    def _packet_handler(self, count, slpack):
        """
        Custom packet handler that accumulates all waveform packets in a
        stream.
        """
        # check if not a complete packet
        if slpack is None or (slpack == SLPacket.SLNOPACKET) or \
                (slpack == SLPacket.SLERROR):
            return False

        # get basic packet info
        type_ = slpack.get_type()
        if self.debug:
            print(type_)

        # process INFO packets here
        if type_ == SLPacket.TYPE_SLINF:
            if self.debug:
                print(SLPacket.TYPE_SLINF)
            return False
        elif type_ == SLPacket.TYPE_SLINFT:
            if self.debug:
                print("Complete INFO:",
                      self._slclient.slconn.get_info_string())
            return True

        # process packet data
        trace = slpack.get_trace()
        if trace is None:
            if self.debug:
                print("Blockette contains no trace")
            return False

        # new samples add to the main stream which is then trimmed
        self.stream += trace
        self.stream.merge(-1)
        return False

    def _fast_packet_handler(self, count, slpack):
        """
        Custom packet handler that accumulates all received data in a
        list of numpy arrays. These can then be converted in one go,
        reducing the processing time up to 90% compared to
        _packet_handler().
        """
        # check if not a complete packet
        if slpack is None or (slpack == SLPacket.SLNOPACKET) or \
                (slpack == SLPacket.SLERROR):
            return False

        # get basic packet info
        type_ = slpack.get_type()
        if self.debug:
            print(type_)

        # process INFO packets here
        if type_ == SLPacket.TYPE_SLINF:
            if self.debug:
                print(SLPacket.TYPE_SLINF)
            return False
        elif type_ == SLPacket.TYPE_SLINFT:
            if self.debug:
                print("Complete INFO:",
                      self._slclient.slconn.get_info_string())
            return True

        # process packet data
        msr_record = slpack.msrecord
        if msr_record is None:
            if self.debug:
                print("Blockette contains no trace")
            return False

        # add to bulk_data_container
        self.bulk_data_container.append(np.frombuffer(msr_record,
                                        dtype=np.int8))
        return False


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests the SeismicPhase class.
"""
from pathlib import Path

import pytest

from obspy.core.util.misc import TemporaryWorkingDirectory
from obspy.taup import TauPyModel
from obspy.taup.tau_model import TauModel
from obspy.taup.seismic_phase import SeismicPhase
from obspy.taup.taup_create import build_taup_model
from obspy.taup.helper_classes import TauModelError


class TestTauPySeismicPhase:
    """
    Test suite for the SeismicPhase class.
    """
    depth = 119

    @pytest.fixture(scope='class')
    def tau_model(self):
        """Return the tau model for testing."""
        return TauModel.from_file('iasp91').depth_correct(self.depth)

    def test_shoot_existing_ray_param(self, tau_model):
        self.do_shoot_existing_ray_param_for_phase('P', tau_model)
        self.do_shoot_existing_ray_param_for_phase('S', tau_model)
        self.do_shoot_existing_ray_param_for_phase('p', tau_model)
        self.do_shoot_existing_ray_param_for_phase('p', tau_model)
        self.do_shoot_existing_ray_param_for_phase('s', tau_model)
        self.do_shoot_existing_ray_param_for_phase('PP', tau_model)
        self.do_shoot_existing_ray_param_for_phase('SS', tau_model)
        self.do_shoot_existing_ray_param_for_phase('PcP', tau_model)
        self.do_shoot_existing_ray_param_for_phase('ScS', tau_model)
        self.do_shoot_existing_ray_param_for_phase('PKP', tau_model)
        self.do_shoot_existing_ray_param_for_phase('SKS', tau_model)
        self.do_shoot_existing_ray_param_for_phase('PKIKP', tau_model)
        self.do_shoot_existing_ray_param_for_phase('SKIKS', tau_model)

    def do_shoot_existing_ray_param_for_phase(self, phase_name, tau_model):
        phase = SeismicPhase(phase_name, tau_model)
        for i, ray_param in enumerate(phase.ray_param):
            max_rp_arrival = phase.shoot_ray(-1, ray_param)
            assert abs(phase.dist[i]-max_rp_arrival.purist_dist) < 0.0001
            assert abs(phase.time[i]-max_rp_arrival.time) < 0.0001

    def test_shoot_middle_ray_param(self, tau_model):
        phase = SeismicPhase('P', tau_model)
        for i in range(phase.ray_param.shape[0] - 1):
            rp = (phase.ray_param[i] + phase.ray_param[i + 1]) / 2
            time_tol = abs(phase.time[i] - phase.time[i + 1])
            max_rp_arrival = phase.shoot_ray(-1, rp)
            assert abs(phase.dist[i]-max_rp_arrival.purist_dist) < 0.1
            assert abs(phase.time[i]-max_rp_arrival.time) < time_tol
            assert abs(phase.dist[i + 1]-max_rp_arrival.purist_dist) < 0.1
            assert abs(phase.time[i + 1]-max_rp_arrival.time) < time_tol

    def test_many_identically_named_phases(self, testdata):
        """
        Regression test to make sure obspy.taup works with models that
        produce many identically names seismic phases.
        """
        with TemporaryWorkingDirectory():
            folder = Path()
            model_name = "smooth_geodynamic_model"
            build_taup_model(
                filename=testdata[model_name + ".tvel"],
                output_folder=folder, verbose=False)
            m = TauPyModel(str(folder / (model_name + ".npz")))
        arr = m.get_ray_paths(172.8000, 46.762440693494824, ["SS"])
        assert len(arr) > 10

    def test_diffracted_phases(self):
        """
        Test of exotic diffracted phases.
        """
        model = TauPyModel('iasp91')
        phs = ["SedPdiffKP", "PdiffPdiff", "PedPdiffKKP",
               "PdiffKKPdiff", "PPdiff", "SKdiffP"]
        dists = [155.0, 210.0, 310.0, 300.0, 220.0, 200.0]
        times = [1464.97, 1697.88, 2052.42, 2008.03, 1742.27, 1431.53]

        for ph, dist, time in zip(phs, dists, times):
            phase = SeismicPhase(ph, model.model)
            arrivals = phase.calc_time(dist)
            arrival = arrivals[0]
            phase.calc_pierce_from_arrival(arrival)
            phase.calc_path_from_arrival(arrival)

            tol = 1e-2
            assert abs(arrival.time - time) < tol
            assert abs(arrival.pierce["time"][-1] - time) < tol
            assert abs(arrival.path["time"][-1] - time) < tol

    def test_phase_names(self, tau_model):
        """
        Simple check to see if illegal phase names are caught.
        """
        legal_phase_names = ["ScS", "ScSScS"]
        illegal_phase_names = ["ScScS", "PKIKPKIKP", "PKIKIKP"]

        for name in legal_phase_names:
            SeismicPhase(name, tau_model)

        for name in illegal_phase_names:
            with pytest.raises(TauModelError):
                SeismicPhase(name, tau_model)



# -*- coding: utf-8 -*-
from .blockette import Blockette
from ..fields import FixedString, Float, Integer, Loop, VariableString
from ..utils import lookup_code, format_resp


RESP = """\
#\t\t+                     +--------------------------------+                 \
     +
#\t\t+                     |   FIR response,%6s ch %s   |                     \
 +
#\t\t+                     +--------------------------------+                 \
     +
#\t\t
B061F03     Stage sequence number:                 %s
B061F05     Symmetry type:                         %s
B061F06     Response in units lookup:              %s - %s
B061F07     Response out units lookup:             %s - %s
B061F08     Number of numerators:                  %s
"""


class Blockette061(Blockette):
    """
    Blockette 061: FIR Response Blockette.

    The FIR blockette is used to specify FIR (Finite Impulse Response) digital
    filter coefficients. It is an alternative to blockette [54] when
    specifying FIR filters. The blockette recognizes the various forms of
    filter symmetry and can exploit them to reduce the number of factors
    specified to the blockette. In July 2007, the FDSN adopted a convention
    that requires the coefficients to be listed in forward time order.
    As a reference, minimum-phase filters (which are asymmetric) should be
    written with the largest values near the beginning of the coefficient list.
    """

    id = 61
    name = "FIR Response"
    fields = [
        Integer(3, "Stage sequence number", 2),
        VariableString(4, "Response Name", 1, 25, 'UN_'),
        FixedString(5, "Symmetry Code", 1, 'U'),
        Integer(6, "Signal In Units", 3, xpath=34),
        Integer(7, "Signal Out Units", 3, xpath=34),
        Integer(8, "Number of Coefficients", 4),
        # REPEAT field 9 for the Number of Coefficients
        Loop("FIR Coefficient", "Number of Coefficients", [
            Float(9, "FIR Coefficient", 14, mask='%+1.7e')], flat=True),
    ]

    def get_resp(self, station, channel, abbreviations):
        """
        Returns RESP string.
        """
        out = RESP % (station, channel,
                      self.stage_sequence_number,
                      self.symmetry_code,
                      lookup_code(abbreviations, 34, 'unit_name',
                                  'unit_lookup_code', self.signal_in_units),
                      lookup_code(abbreviations, 34, 'unit_description',
                                  'unit_lookup_code', self.signal_in_units),
                      lookup_code(abbreviations, 34, 'unit_name',
                                  'unit_lookup_code', self.signal_out_units),
                      lookup_code(abbreviations, 34, 'unit_description',
                                  'unit_lookup_code', self.signal_out_units),
                      self.number_of_coefficients)
        if self.number_of_coefficients > 1:
            out += '#\t\tNumerator coefficients:\n'
            out += '#\t\t  i, coefficient\n'
            for _i in range(self.number_of_coefficients):
                out += 'B061F09    %4s %13s\n' % \
                    (_i, format_resp(self.FIR_coefficient[_i], 6))
        elif self.number_of_coefficients == 1:
            out += '#\t\tNumerator coefficients:\n'
            out += '#\t\t  i, coefficient\n'
            out += 'B061F09    %4s %13s\n' % \
                (0, format_resp(self.FIR_coefficient, 6))
        out += '#\t\t\n'
        return out.encode()



import numpy as np
import matplotlib.pyplot as plt

import obspy
import obspy.signal


st = obspy.read("https://examples.obspy.org/RJOB_061005_072159.ehz.new")
data = st[0].data
npts = st[0].stats.npts
samprate = st[0].stats.sampling_rate

# Filtering the Stream object
st_filt = st.copy()
st_filt.filter('bandpass', freqmin=1, freqmax=3, corners=2, zerophase=True)

# Envelope of filtered data
data_envelope = obspy.signal.filter.envelope(st_filt[0].data)

# The plotting, plain matplotlib
t = np.arange(0, npts / samprate, 1 / samprate)
plt.plot(t, st_filt[0].data, 'k')
plt.plot(t, data_envelope, 'k:')
plt.title(st[0].stats.starttime)
plt.ylabel('Filtered Data w/ Envelope')
plt.xlabel('Time [s]')
plt.xlim(80, 90)
plt.show()



"""
obspy.imaging - Plotting routines for ObsPy
===========================================
This module provides routines for plotting and displaying often used in
seismology. It can currently plot waveform data, generate spectrograms and draw
beachballs. The module :mod:`obspy.imaging` depends on the plotting module
`matplotlib <http://matplotlib.org/>`_.


:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)


Seismograms
-----------
This submodule can plot multiple :class:`~obspy.core.trace.Trace` in one
:class:`~obspy.core.stream.Stream` object and has various other optional
arguments to adjust the plot, such as color and tick format changes.

Additionally the start and end time of the plot can be given as
:class:`~obspy.core.utcdatetime.UTCDateTime` objects.

Examples files may be retrieved via https://examples.obspy.org.

>>> from obspy import read
>>> st = read()
>>> print(st)  # doctest: +ELLIPSIS
3 Trace(s) in Stream:
BW.RJOB..EHZ | 2009-08-24T00:20:03.000000Z - ... | 100.0 Hz, 3000 samples
BW.RJOB..EHN | 2009-08-24T00:20:03.000000Z - ... | 100.0 Hz, 3000 samples
BW.RJOB..EHE | 2009-08-24T00:20:03.000000Z - ... | 100.0 Hz, 3000 samples
>>> st.plot(color='gray', tick_format='%I:%M %p',  # doctest: +ELLIPSIS
...         starttime=st[0].stats.starttime,
...         endtime=st[0].stats.starttime+20)
<...Figure ...>

.. plot::

    from obspy import read
    st = read()
    st.plot(color='gray', tick_format='%I:%M %p',
            starttime=st[0].stats.starttime,
            endtime=st[0].stats.starttime+20)

Spectrograms
------------
The :mod:`obspy.imaging.spectrogram` submodule plots spectrograms.

The spectrogram will on default have 90% overlap and a maximum sliding window
size of 4096 points. For more info see
:func:`obspy.imaging.spectrogram.spectrogram`.

>>> from obspy import read
>>> st = read()
>>> st[0].spectrogram(log=True)  # doctest: +SKIP

.. plot::

    from obspy import read
    st = read()
    st[0].spectrogram(log=True)

Beachballs
----------
Draws a beach ball diagram of an earthquake focal mechanism.

.. note::
    ObsPy ships with two engines for beachball generation.

    (1) :mod:`obspy.imaging.beachball`
        is based on the program from the
        `Generic Mapping Tools (GMT)`_ and the MATLAB script `bb.m`_ written by
        Andy Michael and Oliver Boyd, which both have known limitations.
    (2) :mod:`obspy.imaging.mopad_wrapper`
        is based on the the *Moment tensor Plotting and Decomposition tool*
        (MoPaD) [Krieger2012]_. MoPaD is more correct, however it consumes much
        more processing time.

    The function calls for creating beachballs are similar in both modules. The
    following examples are based on the first module, however those example
    will also work with MoPaD by using

    >>> from obspy.imaging.mopad_wrapper import beachball

    and

    >>> from obspy.imaging.mopad_wrapper import beach

    respectively.

.. _`Generic Mapping Tools (GMT)`: https://www.generic-mapping-tools.org
.. _`bb.m`: http://www.ceri.memphis.edu/people/olboyd/Software/Software.html


.. rubric:: Examples

(1) The focal mechanism can be given by 3 (strike, dip, and rake) components.
    The strike is of the first plane, clockwise relative to north. The dip is
    of the first plane, defined clockwise and perpendicular to strike, relative
    to horizontal such that 0 is horizontal and 90 is vertical. The rake is of
    the first focal plane solution. 90 moves the hanging wall up-dip (thrust),
    0 moves it in the strike direction (left-lateral), -90 moves it down-dip
    (normal), and 180 moves it opposite to strike (right-lateral).

    >>> from obspy.imaging.beachball import beachball
    >>> np1 = [150, 87, 1]
    >>> fig = beachball(np1)  # doctest: +ELLIPSIS

    .. plot::

        from obspy.imaging.beachball import beachball
        np1 = [150, 87, 1]
        beachball(np1)

(2) The focal mechanism can also be specified using the 6 independent
    components of the moment tensor (M11, M22, M33, M12, M13, M23). For
    :func:`obspy.imaging.beachball.beachball` (1, 2, 3) corresponds to
    (Up, South, East) which is equivalent to (r, theta, phi). For
    :func:`obspy.imaging.mopad_wrapper.beachball` the coordinate system can be
    chosen and includes the choices 'NED' (North, East, Down), 'USE' (Up,
    South, East), 'NWU' (North, West, Up) or 'XYZ'.

    >>> from obspy.imaging.beachball import beachball
    >>> mt = [-2.39, 1.04, 1.35, 0.57, -2.94, -0.94]
    >>> fig = beachball(mt) #doctest: +ELLIPSIS

    .. plot::

        from obspy.imaging.beachball import beachball
        mt = [-2.39, 1.04, 1.35, 0.57, -2.94, -0.94]
        beachball(mt)

    For more info see :func:`obspy.imaging.beachball.beachball` and
    :func:`obspy.imaging.mopad_wrapper.beachball`.

(3) Plot the beach ball as matplotlib collection into an existing plot.

    >>> import matplotlib.pyplot as plt
    >>> from obspy.imaging.beachball import beach
    >>>
    >>> np1 = [150, 87, 1]
    >>> mt = [-2.39, 1.04, 1.35, 0.57, -2.94, -0.94]
    >>> beach1 = beach(np1, xy=(-70, 80), width=30)
    >>> beach2 = beach(mt, xy=(50, 50), width=50)
    >>>
    >>> plt.plot([-100, 100], [0, 100], "rv", ms=20) #doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> ax = plt.gca()
    >>> ax.add_collection(beach1) #doctest: +SKIP
    >>> ax.add_collection(beach2) #doctest: +SKIP
    >>> ax.set_aspect("equal")
    >>> ax.set_xlim((-120, 120))  # doctest: +SKIP
    (-120, 120)
    >>> ax.set_ylim((-20, 120))  # doctest: +SKIP
    (-20, 120)

    .. plot::

        import matplotlib.pyplot as plt
        from obspy.imaging.beachball import beach
        np1 = [150, 87, 1]
        mt = [-2.39, 1.04, 1.35, 0.57, -2.94, -0.94]
        beach1 = beach(np1, xy=(-70, 80), width=30)
        beach2 = beach(mt, xy=(50, 50), width=50)
        plt.plot([-100, 100], [0, 100], "rv", ms=20)
        ax = plt.gca()
        ax.add_collection(beach1)
        ax.add_collection(beach2)
        ax.set_aspect("equal")
        ax.set_xlim((-120, 120))
        ax.set_ylim((-20, 120))

    For more info see :func:`obspy.imaging.beachball.beach` and
    :func:`obspy.imaging.mopad_wrapper.beach`.


Saving plots into files
-----------------------
All plotting routines offer an ``outfile`` argument to save the result into a
file.

The ``outfile`` parameter is also used to automatically determine the file
format. Available output formats mainly depend on your matplotlib settings.
Common formats are png, svg, pdf or ps.

>>> from obspy import read
>>> st = read()
>>> st.plot(outfile='graph.png') #doctest: +SKIP
"""
# Please do not import any modules using matplotlib - otherwise it will disturb
# the test suite (running without X11 or any other display)
# see tests/__init__.py for details



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: rotate.py
#  Purpose: Various Seismogram Rotation Functions
#   Author: Tobias Megies, Tom Richter, Lion Krischer
#    Email: tobias.megies@geophysik.uni-muenchen.de
#
# Copyright (C) 2009-2013 Tobias Megies, Tom Richter, Lion Krischer
# --------------------------------------------------------------------
"""
Various Seismogram Rotation Functions

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import warnings
from math import cos, sin, radians

import numpy as np


def rotate_ne_rt(n, e, ba):
    """
    Rotates horizontal components of a seismogram.

    The North- and East-Component of a seismogram will be rotated in Radial
    and Transversal Component. The angle is given as the back-azimuth, that is
    defined as the angle measured between the vector pointing from the station
    to the source and the vector pointing from the station to the North.

    :type n: :class:`~numpy.ndarray`
    :param n: Data of the North component of the seismogram.
    :type e: :class:`~numpy.ndarray`
    :param e: Data of the East component of the seismogram.
    :type ba: float
    :param ba: The back azimuth from station to source in degrees.
    :return: Radial and Transversal component of seismogram.
    """
    if len(n) != len(e):
        raise TypeError("North and East component have different length.")
    if ba < 0 or ba > 360:
        raise ValueError("Back Azimuth should be between 0 and 360 degrees.")
    ba = radians(ba)
    r = - e * sin(ba) - n * cos(ba)
    t = - e * cos(ba) + n * sin(ba)
    return r, t


def rotate_rt_ne(r, t, ba):
    """
    Rotates horizontal components of a seismogram.

    Rotates from radial and transversal components to North and East
    components.

    This is the inverse transformation of the transformation described
    in :func:`rotate_ne_rt`.

    :type r: :class:`~numpy.ndarray`
    :param r: Data of the Radial component of the seismogram.
    :type t: :class:`~numpy.ndarray`
    :param t: Data of the Transverse component of the seismogram.
    :returns: North and East component of seismogram.
    """
    ba = 360.0 - ba
    return rotate_ne_rt(r, t, ba)


def rotate_zne_lqt(z, n, e, ba, inc):
    """
    Rotates all components of a seismogram.

    The components will be rotated from ZNE (Z, North, East, left-handed) to
    LQT (e.g. ray coordinate system, right-handed). The rotation angles are
    given as the back-azimuth and inclination.

    The transformation consists of 3 steps::

        1. mirroring of E-component at ZN plain: ZNE -> ZNW
        2. negative rotation of coordinate system around Z-axis with angle ba:
           ZNW -> ZRT
        3. negative rotation of coordinate system around T-axis with angle inc:
           ZRT -> LQT

    :type z: :class:`~numpy.ndarray`
    :param z: Data of the Z component of the seismogram.
    :type n: :class:`~numpy.ndarray`
    :param n: Data of the North component of the seismogram.
    :type e: :class:`~numpy.ndarray`
    :param e: Data of the East component of the seismogram.
    :type ba: float
    :param ba: The back azimuth from station to source in degrees.
    :type inc: float
    :param inc: The inclination of the ray at the station in degrees.
    :return: L-, Q- and T-component of seismogram.
    """
    if len(z) != len(n) or len(z) != len(e):
        raise TypeError("Z, North and East component have different length!?!")
    if ba < 0 or ba > 360:
        raise ValueError("Back Azimuth should be between 0 and 360 degrees!")
    if inc < 0 or inc > 360:
        raise ValueError("Inclination should be between 0 and 360 degrees!")
    ba = radians(ba)
    inc = radians(inc)
    l = z * cos(inc) - n * sin(inc) * cos(ba) - e * sin(inc) * sin(ba)  # NOQA
    q = z * sin(inc) + n * cos(inc) * cos(ba) + e * cos(inc) * sin(ba)  # NOQA
    t = n * sin(ba) - e * cos(ba)  # NOQA
    return l, q, t


def rotate_lqt_zne(l, q, t, ba, inc):  # NOQA
    """
    Rotates all components of a seismogram.

    The components will be rotated from LQT to ZNE.
    This is the inverse transformation of the transformation described
    in :func:`rotate_zne_lqt`.
    """
    if len(l) != len(q) or len(l) != len(t):
        raise TypeError("L, Q and T component have different length!?!")
    if ba < 0 or ba > 360:
        raise ValueError("Back Azimuth should be between 0 and 360 degrees!")
    if inc < 0 or inc > 360:
        raise ValueError("Inclination should be between 0 and 360 degrees!")
    ba = radians(ba)
    inc = radians(inc)
    z = l * cos(inc) + q * sin(inc)
    n = -l * sin(inc) * cos(ba) + q * cos(inc) * cos(ba) + t * sin(ba)
    e = -l * sin(inc) * sin(ba) + q * cos(inc) * sin(ba) - t * cos(ba)
    return z, n, e


def _dip_azimuth2zne_base_vector(dip, azimuth):
    """
    Helper function converting a vector described with azimuth and dip of unit
    length to a vector in the ZNE (Vertical, North, East) base.

    The definition of azimuth and dip is according to the SEED reference
    manual.
    """
    dip = np.deg2rad(dip)
    azimuth = np.deg2rad(azimuth)

    return np.array([-np.sin(dip),
                     np.cos(azimuth) * np.cos(dip),
                     np.sin(azimuth) * np.cos(dip)])


def rotate2zne(data_1, azimuth_1, dip_1, data_2, azimuth_2, dip_2, data_3,
               azimuth_3, dip_3, inverse=False):
    """
    Rotates an arbitrarily oriented three-component vector to ZNE.

    Each components orientation is described with a azimuth and a dip. The
    azimuth is defined as the degrees from North, clockwise and the dip is the
    defined as the number of degrees, down from horizontal. Both definitions
    are according to the SEED standard.

    The three components need not be orthogonal to each other but the
    components have to be linearly independent. The function performs a full
    base change to orthogonal Vertical, North, and East orientations.

    :param data_1: Data component 1.
    :param azimuth_1: The azimuth of component 1.
    :param dip_1: The dip of component 1.
    :param data_2: Data component 2.
    :param azimuth_2: The azimuth of component 2.
    :param dip_2: The dip of component 2.
    :param data_3: Data component 3.
    :param azimuth_3: The azimuth of component 3.
    :param dip_3: The dip of component 3.
    :param inverse: If `True`, the data arrays will be converted from ZNE to
        whatever coordinate system the azimuths and dips specify. In that
        case data_1, data_2, data_3 have to be data arrays for Z, N,
        and E and the dips and azimuths specify where to transform to.
    :type inverse: bool

    :rtype: tuple(:class:`numpy.ndarray`, :class:`numpy.ndarray`,
        :class:`numpy.ndarray`)
    :returns: The three rotated components, oriented in Z, N, and E if
        `inverse` is `False`. Otherwise they will be oriented as specified
        by the dips and azimuths.

    An input of ZNE yields an output of ZNE

    >>> rotate2zne(np.arange(3), 0, -90, np.arange(3) * 2, 0, 0, \
            np.arange(3) * 3, 90, 0) # doctest: +NORMALIZE_WHITESPACE
    (array([ 0., 1., 2.]), array([ 0., 2., 4.]), array([ 0., 3., 6.]))

    An input of ZSE yields an output of ZNE

    >>> rotate2zne(np.arange(3), 0, -90, np.arange(3) * 2, 180, 0, \
            np.arange(3) * 3, 90, 0) # doctest: +NORMALIZE_WHITESPACE
    (array([ 0., 1., 2.]), array([ 0., -2., -4.]), array([ 0., 3., 6.]))

    Mixed up components should get rotated to ZNE.

    >>> rotate2zne(np.arange(3), 0, 0, np.arange(3) * 2, 90, 0, \
            np.arange(3) * 3, 0, -90) # doctest: +NORMALIZE_WHITESPACE
    (array([ 0., 3., 6.]), array([ 0., 1., 2.]), array([ 0., 2., 4.]))
    """
    if len(set(len(i_) for i_ in (data_1, data_2, data_3))) != 1:
        msg = "All three data arrays must be of same length."
        raise ValueError(msg)

    # Define the base vectors of the old base in terms of the new base vectors.
    base_vector_1 = _dip_azimuth2zne_base_vector(dip_1, azimuth_1)
    base_vector_2 = _dip_azimuth2zne_base_vector(dip_2, azimuth_2)
    base_vector_3 = _dip_azimuth2zne_base_vector(dip_3, azimuth_3)

    # Base change matrix.
    m = np.array([base_vector_1,
                  base_vector_2,
                  base_vector_3])

    # Determinant gives the volume change of a unit cube going from one
    # basis to the next. It should neither be too small nor to large. These
    # here are arbitrary limits.
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore',
                                '.*invalid value encountered in det.*')
        det = np.linalg.det(m)
    if not (1E-6 < abs(det) < 1E6):
        raise ValueError("The given directions are not linearly independent, "
                         "at least within numerical precision. Determinant "
                         "of the base change matrix: %g" % det)

    if not inverse:
        m = np.linalg.inv(m)

    z, n, e = np.dot(m, [data_1, data_2, data_3])

    # Replace all negative zeros. These might confuse some further
    # processing programs.
    z = np.array(z).ravel()
    z[z == -0.0] = 0
    n = np.array(n).ravel()
    n[n == -0.0] = 0
    e = np.array(e).ravel()
    e[e == -0.0] = 0

    return z, n, e


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The obspy.io.dmx.core test suite.
"""
import numpy as np

from obspy import read
from obspy.core.utcdatetime import UTCDateTime
from obspy.io.dmx.core import _read_dmx


class TestCore():
    """
    Test cases for dmx core interface
    """
    def test_read_via_obspy(self, testdata):
        """
        Read files via obspy.core.stream.read function.
        """
        filename = testdata['131114_090600.dmx']
        # 1
        st = read(filename)
        st.verify()
        assert len(st) == 2
        assert st[0].stats.starttime == UTCDateTime(2013, 11, 14, 9, 6)
        assert st[0].stats.endtime == \
            UTCDateTime(2013, 11, 14, 9, 6, 59, 990000)
        assert "dmx" in st[0].stats
        assert len(st[0]) == 6000
        assert round(abs(st[0].stats.sampling_rate-100.0), 7) == 0
        assert st[0].stats.channel == 'Z'
        assert st[0].id == 'ETNA.EMFO..Z'

    def test_read_via_module(self, testdata):
        """
        Read files via obspy.io.mdx.core._read_dmx function directly.
        """
        filename = testdata['131114_090600.dmx']
        # 1
        st = _read_dmx(filename)
        st.verify()
        assert len(st) == 2
        assert st[0].stats.starttime == UTCDateTime(2013, 11, 14, 9, 6)
        assert st[0].stats.endtime == \
            UTCDateTime(2013, 11, 14, 9, 6, 59, 990000)
        assert "dmx" in st[0].stats
        assert len(st[0]) == 6000
        assert round(abs(st[0].stats.sampling_rate-100.0), 7) == 0
        assert st[0].stats.channel == 'Z'
        assert st[0].id == 'ETNA.EMFO..Z'

    def test_read_with_station(self, testdata):
        """
        Read files and passing a station keyword argument.
        """
        filename = testdata['131114_090600.dmx']
        # 1
        st = read(filename, station='EMPL')
        st.verify()

        assert len(st) == 1
        assert st[0].id == "ETNA.EMPL..Z"
        for tr in st:
            assert tr.stats.station == "EMPL"

    def test_check_data_content_sum(self, testdata):
        """
        Read files and passing a station keyword argument.
        """
        filename = testdata['131114_090600.dmx']
        # 1
        st = read(filename)
        st.verify()
        assert st[0].data.sum() == -90928.0
        np.testing.assert_array_equal(
            st[0].data[:5], [148., 238., 133., 9., -92.])

        assert st[1].data.sum() == 3003120.0
        np.testing.assert_array_equal(
            st[1].data[:5], [537., 721., 844., 924., 977.])
